from __future__ import annotations

import typing

import toolsql

from ctc import config
from ctc import evm
from ctc import spec
from ctc import db

from . import chainlink_schema_defs
from . import chainlink_statements


if typing.TYPE_CHECKING:

    import toolsql

    RawChainlinkFeed = typing.Mapping[typing.Any, typing.Any]

    ChainlinkFeedPayload = typing.Mapping[
        'ChainlinkNetworkName',
        'ChainlinkNetworkGroup',
    ]
    ChainlinkNetworkName = str

    class ChainlinkNetworkGroup(typing.TypedDict):
        title: str
        feedType: str
        networks: list['ChainlinkNetworkData']

    class ChainlinkNetworkData(typing.TypedDict):
        name: str
        url: str
        networkType: str
        proxies: list[ChainlinkProxyData]

    class ChainlinkProxyData(typing.TypedDict):
        pair: str
        assetName: str
        deviationThreshold: int
        heartbeat: str
        decimals: int
        proxy: str
        feedCategory: str
        feedType: str


# # legacy version of chainlink payload data
# _legacy_network_payload_locations: typing.Mapping[spec.NetworkName, tuple[str, str]] = {
#     'ethereum': ('ethereum-addresses', 'Ethereum Mainnet'),
#     'kovan': ('ethereum-addresses', 'Kovan Testnet'),
#     'rinkeby': ('ethereum-addresses', 'Rinkeby Testnet'),
#     'bnb': ('bnb-chain-addresses-price', 'BNB Chain Mainnet'),
#     'bnb_testnet': ('bnb-chain-addresses-price', 'BNB Chain Testnet'),
#     'polygon': ('matic-addresses', 'Polygon Mainnet'),
#     'polygon_mumbai': ('matic-addresses', 'Mumbai Testnet'),
#     'gnosis': ('data-feeds-gnosis-chain', 'Gnosis Chain Mainnet'),
#     'heco': ('huobi-eco-chain-price-feeds', 'HECO Mainnet'),
#     'avalanche': ('avalanche-price-feeds', 'Avalanche Mainnet'),
#     'avalanche_fuji': ('avalanche-price-feeds', 'Avalanche Testnet'),
#     'fantom': ('fantom-price-feeds', 'Fantom Mainnet'),
#     'arbitrum': ('arbitrum-price-feeds', 'Arbitrum Mainnet'),
#     'arbitrum_rinkeby': ('arbitrum-price-feeds', 'Arbitrum Rinkeby'),
#     'harmony': ('harmony-price-feeds', 'Harmony Mainnet'),
#     'harmony_testnet': ('harmony-price-feeds', 'Harmony Testnet'),
#     # 'solana': ('data-feeds-solana', 'Solana Mainnet'),
#     # 'solana_testnet': ('data-feeds-solana', 'Solana Devnet'),
#     'optimism': ('optimism-price-feeds', 'Optimism Mainnet'),
#     'optimism_kovan': ('optimism-price-feeds', 'Optimism Kovan'),
#     'moonriver': ('data-feeds-moonriver', 'Moonriver Mainnet'),
#     'moonbeam': ('data-feeds-moonbeam', 'Moonbeam Mainnet'),
# }

network_payload_locations: typing.Mapping[spec.NetworkName, tuple[str, int]] = {
    'ethereum': ('ethereum', 0),
    'goerli': ('ethereum', 1),
    'bnb': ('bnb-chain', 0),
    'bnb_testnet': ('bnb-chain', 1),
    'polygon': ('polygon', 0),
    'polygon_mumbai': ('polygon', 0),
    'gnosis': ('gnosis-chain', 0),
    'heco': ('heco-chain', 0),
    'avalanche': ('avalanche', 0),
    'avalanche_fuji': ('avalanche', 1),
    'fantom': ('fantom', 0),
    'fantom_testnet': ('fantom', 1),
    'arbitrum': ('arbitrum', 0),
    'arbitrum_rinkeby': ('arbitrum', 1),
    'harmony': ('harmony', 0),
    'optimism': ('optimism', 0),
    'optimism_goerli': ('optimism', 1),
    'moonriver': ('moonriver', 0),
    'moonbeam': ('moonbeam', 0),
    'metis': ('metis', 0),
    'klaytn': ('klaytn', 0),
}


async def async_get_complete_feed_payload() -> ChainlinkFeedPayload:

    import aiohttp

    url = 'https://cl-docs-addresses.web.app/addresses.json'

    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            return await response.json()  # type: ignore


async def async_get_network_feed_data(
    network: spec.NetworkReference,
    payload: ChainlinkFeedPayload | None = None,
) -> typing.Sequence[ChainlinkProxyData]:
    if payload is None:
        payload = await async_get_complete_feed_payload()

    if network in network_payload_locations:
        if isinstance(network, int):
            network = evm.get_network_chain_id(network)
        if not isinstance(network, str):
            raise Exception('unknown network type: ' + str(type(network)))
        network_group, index = network_payload_locations[network]
    else:
        raise Exception('unknown network')

    return payload[network_group]['networks'][index]['proxies']


async def async_import_networks_to_db(
    networks: typing.Sequence[ChainlinkNetworkName] | None = None,
    *,
    db_config: toolsql.DBConfig,
    payload: ChainlinkFeedPayload | None = None,
    verbose: bool = True,
) -> None:
    """import multiple networks of feeds to db

    by default import all networks
    """

    if payload is None:
        payload = await async_get_complete_feed_payload()

    # determine which networks to use
    if networks is None:

        known_networks = {
            datum['name'] for datum in evm.get_networks().values()
        }

        networks = []
        for network_name in network_payload_locations.keys():
            network_group, index = network_payload_locations[network_name]
            if (
                network_name in known_networks
                and network_group in payload
                and len(payload[network_group]['networks']) > index
            ):
                networks.append(network_name)

    #         # use all networks
    #         locations_to_network = {
    #             v: k for k, v in network_payload_locations.items()
    #         }
    #         networks = []
    #         for key in payload.keys():
    #             for network_data in payload[key]['networks']:
    #                 subkey = network_data['name']
    #                 if (key, subkey) in locations_to_network:
    #                     network = locations_to_network[(key, subkey)]
    #                     networks.append(network)

    if verbose:
        import toolstr
        from ctc import cli

        styles = cli.get_cli_styles()

        toolstr.print(
            'Adding Chainlink feed metadata to db for',
            toolstr.add_style(
                str(len(networks)), styles['description'] + ' bold'
            ),
            'networks...',
        )

    # add each network
    for network in networks:

        # create table
        table_schema = db.get_table_schema(
            'chainlink_feeds', context={'network': network}
        )
        with toolsql.connect(db_config) as conn:
            toolsql.create_table(
                table_schema,
                if_not_exists=True,
                conn=conn,
                confirm=True,
            )

        # insert feed rows
        async with toolsql.async_connect(db_config) as conn:
            await async_import_network_to_db(
                network=network,
                payload=payload,
                conn=conn,
                verbose=verbose,
                indent=4,
            )


async def async_import_network_to_db(
    network: ChainlinkNetworkName,
    *,
    payload: ChainlinkFeedPayload | None = None,
    conn: toolsql.AsyncConnection,
    verbose: bool = True,
    indent: int | str | None = None,
) -> None:

    raw_feeds = await async_get_network_feed_data(
        network=network, payload=payload
    )

    rename_fields = {
        'proxy': 'address',
        'pair': 'name',
        'deviationThreshold': 'deviation',
        'heartbeat': 'heartbeat',
        'decimals': 'decimals',
        'assetName': 'asset',
        'feedType': 'asset_type',
        'feedCategory': 'status',
    }

    feeds: typing.Sequence[typing.Mapping[str, typing.Any]] = [
        {key: raw_feed[raw_key] for raw_key, key in rename_fields.items()}  # type: ignore
        for raw_feed in raw_feeds
    ]

    await chainlink_statements.async_upsert_feeds(
        feeds=feeds,
        context=dict(network=network),
        conn=conn,
    )

    if verbose:
        import toolstr
        from ctc import cli

        if indent is None:
            indent = ''
        if isinstance(indent, int):
            indent = ' ' * indent

        styles = cli.get_cli_styles()

        toolstr.print(
            indent + 'added',
            toolstr.add_style(str(len(feeds)), styles['description'] + ' bold'),
            toolstr.add_style(network, styles['metavar']),
            'Chainlink feeds to db',
        )


def print_payload_summary(payload: ChainlinkFeedPayload) -> None:
    for group in payload.keys():
        print(group)
        for network in payload[group]['networks']:
            print(
                '    -',
                network['name'],
                '(' + str(len(network['proxies'])) + ')',
            )


async def async_intake_aggregator_update(
    *,
    feed: spec.Address,
    aggregator: spec.Address,
    block_number: int,
    context: spec.Context,
) -> None:

    db_config = config.get_context_db_config(
        schema_name='chainlink',
        context=context,
    )
    async with toolsql.async_connect(db_config) as conn:
        await chainlink_statements.async_upsert_aggregator_update(
            feed=feed,
            aggregator=aggregator,
            block_number=block_number,
            context=context,
            conn=conn,
        )


async def async_intake_aggregator_updates(
    updates: typing.Sequence[chainlink_schema_defs._FeedAggregatorUpdate],
    context: spec.Context,
) -> None:

    db_config = config.get_context_db_config(
        schema_name='chainlink',
        context=context,
    )
    async with toolsql.async_connect(db_config) as conn:
        await chainlink_statements.async_upsert_aggregator_updates(
            updates=updates,
            conn=conn,
            context=context,
        )




from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import types


class AsyncContextManager:
    """performs cleanup before eventloop is closed"""

    async def __aenter__(self) -> None:
        pass

    async def __aexit__(
        self,
        exception_type: typing.Type[BaseException] | None,
        exception: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        from ctc import rpc

        # TODO: close all sessions, not just default session
        # TODO: close pending db connections
        await rpc.async_close_http_session()



from __future__ import annotations

import typing

from ctc import evm
from ctc import rpc
from ctc import spec

from . import balancer_spec

if typing.TYPE_CHECKING:
    import tooltime


async def async_get_pool_id(
    pool_address: spec.Address,
    block: spec.BlockNumberReference | None = None,
    *,
    context: spec.Context = None,
) -> str:

    function_abi: spec.FunctionABI = {
        'inputs': [],
        'name': 'getPoolId',
        'outputs': [
            {
                'internalType': 'bytes32',
                'name': '',
                'type': 'bytes32',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    }

    result = await rpc.async_eth_call(
        to_address=pool_address,
        function_abi=function_abi,
        block_number=block,
        context=context,
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


async def async_get_pool_address(
    pool_id: str,
    block: spec.BlockNumberReference | None = None,
    *,
    context: spec.Context = None,
) -> spec.Address:

    vault = balancer_spec.vault

    pool = await rpc.async_eth_call(
        to_address=vault,
        function_abi=balancer_spec.vault_function_abis['getPool'],
        function_parameters=[pool_id],
        block_number=block,
        context=context,
    )
    address = pool[0]
    if not isinstance(address, str):
        raise Exception('invalid rpc result')
    return address


async def async_get_pool_tokens(
    *,
    pool_address: spec.Address | None = None,
    pool_id: str | None = None,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> list[spec.Address]:

    vault = balancer_spec.vault
    if pool_id is None:
        if pool_address is None:
            raise Exception('must specify pool_id or pool_address')
        pool_id = await async_get_pool_id(pool_address, context=context)

    pool_tokens = await rpc.async_eth_call(
        to_address=vault,
        function_abi=balancer_spec.vault_function_abis['getPoolTokens'],
        function_parameters=[pool_id],
        block_number=block,
        package_named_outputs=True,
        context=context,
    )
    return list(pool_tokens['tokens'])


async def async_get_token_registrations(
    factory: spec.Address,
    *,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    context: spec.Context = None,
) -> typing.Mapping[str, typing.Sequence[str]]:

    balancer_token_registrations = await evm.async_get_events(
        factory,
        event_abi=balancer_spec.vault_event_abis['TokensRegistered'],
        verbose=False,
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        context=context,
    )

    token_registrations_by_pool: typing.MutableMapping[
        str, typing.MutableSequence[str]
    ] = {}
    for row in balancer_token_registrations.to_dicts():
        pool = row['arg__poolId']

        token_registrations_by_pool.setdefault(pool, [])
        token_registrations_by_pool[pool].extend(row['arg__tokens'])

    return token_registrations_by_pool




from __future__ import annotations

import toolsql


block_timestamps_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'block_timestamps': {
            'columns': [
                {
                    'name': 'block_number',
                    'type': 'Integer',
                    'primary': True,
                },
                {
                    'name': 'timestamp',
                    'type': 'Integer',
                    'index': True,
                    'nullable': False,
                },
            ],
        },
    },
}



from __future__ import annotations

import typing

import toolsql

from ctc import config
from ctc import evm
from ctc import spec
from ... import management
from . import events_statements


async def async_intake_encoded_events(
    *,
    encoded_events: typing.Sequence[spec.EncodedEvent],
    query: spec.DBEventQuery,
    context: spec.Context,
    latest_block: int | None = None,
) -> None:

    import numpy as np

    if len(encoded_events) == 0:
        return

    # only insert blocks after a given number of confirmations
    if latest_block is None:
        latest_block = await evm.async_get_latest_block_number()
    required_confirmations = management.get_required_confirmations(
        context=context
    )
    latest_allowed_block = latest_block - required_confirmations
    if query['start_block'] > latest_allowed_block:
        return
    if query['end_block'] > latest_allowed_block:
        if isinstance(encoded_events[0], dict):
            blocks = np.array([event['block_number'] for event in encoded_events])
            confirmed_mask = blocks <= latest_allowed_block
            encoded_events = [
                event
                for event, confirmed in zip(encoded_events, confirmed_mask)
                if confirmed
            ]
        elif isinstance(encoded_events[0], tuple):
            encoded_events = [
                event
                for event in encoded_events
                if event[0] <= latest_allowed_block
            ]
        else:
            raise Exception()
        query = dict(query, end_block=latest_allowed_block)  # type: ignore

    # return early if no events to intake
    if len(encoded_events) == 0:
        return

    # insert into database
    db_config = config.get_context_db_config(
        schema_name='events',
        context=context,
    )
    async with toolsql.async_connect(db_config) as conn:
        await events_statements.async_upsert_event_query(
            event_query=query,
            conn=conn,
            context=context,
        )
        await events_statements.async_upsert_events(
            encoded_events=encoded_events,
            conn=conn,
            context=context,
        )

    return None




from __future__ import annotations

import typing

from ctc import spec
from ctc import evm
from .. import rpc_request


def construct_trace_transaction(
    transaction_hash: str,
) -> spec.RpcSingularRequest:
    return rpc_request.create(
        'trace_transaction',
        [transaction_hash],
    )


def construct_trace_replay_transaction(
    transaction_hash: str,
    trace_type: typing.Sequence[spec.TraceOutputType],
) -> spec.RpcSingularRequest:
    return rpc_request.create(
        'trace_replayTransaction',
        [transaction_hash, trace_type],
    )


def construct_trace_raw_transaction(
    call_data: str,
    trace_type: typing.Sequence[spec.TraceOutputType],
) -> spec.RpcSingularRequest:
    return rpc_request.create(
        'trace_rawTransaction',
        [call_data, trace_type],
    )


#
# # calls
#


def construct_trace_call(
    to_address: spec.BinaryData,
    *,
    from_address: spec.BinaryData | None = None,
    gas: spec.BinaryData | None = None,
    gas_price: spec.BinaryData | None = None,
    value_sent: spec.BinaryData | None = None,
    call_data: spec.BinaryData | None = None,
    function_parameters: typing.Sequence[typing.Any]
    | typing.Mapping[str, typing.Any]
    | None = None,
    function_abi: spec.FunctionABI | None = None,
    block_number: spec.BlockNumberReference | None = None,
    trace_type: typing.Sequence[spec.TraceOutputType],
) -> spec.RpcSingularRequest:
    from . import rpc_state_constructors

    call = rpc_state_constructors.construct_eth_call(
        to_address=to_address,
        from_address=from_address,
        gas=gas,
        gas_price=gas_price,
        value_sent=value_sent,
        block_number=block_number,
        call_data=call_data,
        function_parameters=function_parameters,
        function_abi=function_abi,
    )
    call['method'] = 'trace_call'
    tx, block_number = call['params']
    call['params'] = [tx, trace_type, block_number]
    return call


def construct_trace_call_many(
    calls: typing.Sequence[typing.Mapping[str, typing.Any]],
    trace_type: typing.Sequence[spec.TraceOutputType] | None,
    *,
    block_number: spec.BlockNumberReference | None = None,
) -> spec.RpcSingularRequest:
    """not an efficient implementation"""

    subrequests = []
    for call in calls:
        if 'block_number' in call:
            raise Exception('specify block_number as a top-level parameter')

        if call.get('trace_type') is not None:
            subrequest = construct_trace_call(**call)
        elif trace_type is not None:
            subrequest = construct_trace_call(trace_type=trace_type, **call)
        else:
            raise Exception('must specify trace_type')

        # parse out call
        sub_call, sub_trace_type, _ = subrequest['params']
        subrequests.append([sub_call, sub_trace_type])

    return rpc_request.create(
        'trace_callMany',
        [subrequests, block_number],
    )


#
# # trace ranges
#


def construct_trace_get(
    transaction_hash: str,
    trace_indices: typing.Sequence[int],
) -> spec.RpcSingularRequest:
    return rpc_request.create(
        'trace_get',
        [transaction_hash, trace_indices],
    )


def construct_trace_filter(
    *,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    from_addresses: spec.Address | None = None,
    to_addresses: spec.Address | None = None,
    after: int | None = None,
    count: int | None = None,
) -> spec.RpcSingularRequest:
    payload = {
        'start_block': start_block,
        'end_block': end_block,
        'from_addresses': from_addresses,
        'to_addresses': to_addresses,
        'after': after,
        'count': count,
    }
    payload = {k: v for k, v in payload.items() if v is not None}
    return rpc_request.create(
        'trace_filter',
        [payload],
    )


def construct_trace_block(
    block_number: spec.BlockNumberReference,
) -> spec.RpcSingularRequest:
    block_number = evm.encode_block_number(block_number)
    return rpc_request.create(
        'trace_block',
        [block_number],
    )


def construct_trace_replay_block_transactions(
    block_number: spec.BlockNumberReference,
    trace_type: typing.Sequence[spec.TraceOutputType] | None,
) -> spec.RpcSingularRequest:
    block_number = evm.encode_block_number(block_number)
    return rpc_request.create(
        'trace_replayBlockTransactions',
        [block_number, trace_type],
    )


#
# # debug traces
#


def construct_debug_trace_call(
    to_address: spec.BinaryData,
    *,
    from_address: spec.BinaryData | None = None,
    gas: spec.BinaryData | None = None,
    gas_price: spec.BinaryData | None = None,
    value_sent: spec.BinaryData | None = None,
    call_data: spec.BinaryData | None = None,
    function_parameters: typing.Sequence[typing.Any]
    | typing.Mapping[str, typing.Any]
    | None = None,
    function_abi: spec.FunctionABI | None = None,
    block_number: spec.BlockNumberReference | None = None,
    trace_type: spec.TraceOutputType | None,
) -> spec.RpcSingularRequest:
    from . import rpc_state_constructors

    call = rpc_state_constructors.construct_eth_call(
        to_address=to_address,
        from_address=from_address,
        gas=gas,
        gas_price=gas_price,
        value_sent=value_sent,
        block_number=block_number,
        call_data=call_data,
        function_parameters=function_parameters,
        function_abi=function_abi,
    )
    call['method'] = 'debug_traceCall'
    tx, block_number = call['params']
    call['params'] = [tx, block_number, trace_type]
    return call


def construct_debug_trace_call_many(
    calls: typing.Sequence[typing.Mapping[str, typing.Any]],
    trace_type: spec.TraceOutputType | None,
    *,
    block_number: spec.BlockNumberReference | None = None,
) -> spec.RpcSingularRequest:
    """not an efficient implementation"""

    subrequests = []
    for call in calls:
        if 'block_number' in call:
            raise Exception('specify block_number as a top-level parameter')

        if call.get('trace_type') is not None:
            subrequest = construct_debug_trace_call(**call)
        subrequest = construct_debug_trace_call(
            trace_type=trace_type,
            block_number=block_number,
            **call,
        )

        # parse out call
        sub_call, block_number, sub_trace_type = subrequest['params']
        subrequests.append([sub_call, sub_trace_type])

    return rpc_request.create(
        'debug_traceCallMany',
        [subrequests, block_number],
    )


def construct_debug_trace_transaction(
    transaction_hash: str,
    trace_type: spec.TraceOutputType | None = None,
) -> spec.RpcSingularRequest:
    return rpc_request.create(
        'debug_traceTransaction',
        [transaction_hash, trace_type],
    )


def construct_debug_trace_block_by_number(
    block_number: spec.BlockNumberReference,
    trace_type: spec.TraceOutputType | None = None,
) -> spec.RpcSingularRequest:
    block_number = evm.encode_block_number(block_number)
    return rpc_request.create(
        'debug_traceBlockByNumber',
        [block_number, trace_type],
    )


def construct_debug_trace_block_by_hash(
    block_hash: str,
    trace_type: spec.TraceOutputType | None = None,
) -> spec.RpcSingularRequest:
    return rpc_request.create(
        'debug_traceBlockByHash',
        [block_hash, trace_type],
    )




from __future__ import annotations

import typing

from .. import abi_utils
from .. import block_utils
from ctc import spec


async def async_get_event_timestamps(
    events: spec.DataFrame,
    *,
    context: spec.Context = None,
) -> typing.Sequence[int]:
    """get timestamps of an events dataframe"""

    # get block_numbers
    block_numbers = typing.cast(
        typing.Sequence[int],
        events['block_number'].to_list(),
    )

    # get timestamps
    return await block_utils.async_get_block_timestamps(
        block_numbers,
        context=context,
    )


async def _async_get_event_names_column(
    events: spec.DataFrame,
    *,
    share_abis_across_contracts: bool = True,
    context: spec.Context = None,
) -> typing.Sequence[str]:

    contract_addresses = events['contract_addresses'].to_list()
    event_hashes = events['event_hash'].to_list()

    # compile which event abis are needed
    event_abi_queries = {}
    for contract_address, event_hash in zip(contract_addresses, event_hashes):

        if share_abis_across_contracts:
            key = event_hash
        else:
            key = (contract_address, event_hash)

        if key not in event_abi_queries:
            event_abi_queries[key] = {
                'contract_address': contract_address,
                'event_hash': event_hash,
            }

    # get event abis
    event_abis = {}
    for key, query in event_abi_queries.items():
        event_abis[key] = await abi_utils.async_get_event_abi(
            context=context, **query
        )

    # construct column
    event_abi_column = []
    for contract_address, event_hash in zip(contract_addresses, event_hashes):

        if share_abis_across_contracts:
            key = event_hash
        else:
            key = (contract_address, event_hash)

        event_abi_column.append(event_abis[key].get('name', ''))

    return event_abi_column




from __future__ import annotations

import typing

from ctc import spec
from . import contract_abi_summary


def get_contract_abi_diff(
    first_contract_abi: spec.ContractABI,
    second_contract_abi: spec.ContractABI,
) -> typing.Mapping[str, spec.ContractABI]:
    """return difference between two contract ABI's"""

    first_by_selectors = contract_abi_summary.get_contract_abi_by_selectors(
        first_contract_abi
    )
    second_by_selectors = contract_abi_summary.get_contract_abi_by_selectors(
        second_contract_abi
    )

    mutual = []
    first_only = []
    second_only = []

    for key, value in first_by_selectors.items():
        if key in second_by_selectors:
            mutual.append(value)
        else:
            first_only.append(value)
    for key, value in second_by_selectors.items():
        if key not in first_by_selectors:
            second_only.append(value)

    return {
        'mutual': mutual,
        'first_only': first_only,
        'second_only': second_only,
    }


def print_contract_abi_diff(
    first_contract_abi: spec.ContractABI,
    second_contract_abi: spec.ContractABI,
    *,
    first_name: str | None = None,
    second_name: str | None = None,
    functions_only: bool = False,
    events_only: bool = False,
) -> None:
    """print difference between contract ABI's"""

    if first_name is None:
        first_name = 'First Contract'
    if second_name is None:
        second_name = 'Second Contract'

    diff = get_contract_abi_diff(first_contract_abi, second_contract_abi)

    if functions_only or not events_only:

        contract_abi_summary.print_contract_abi_functions(
            diff['mutual'],
            title='Mutual Functions',
        )

        print()
        print()
        contract_abi_summary.print_contract_abi_functions(
            diff['first_only'],
            title=first_name + ' Functions',
        )

        print()
        print()
        contract_abi_summary.print_contract_abi_functions(
            diff['second_only'],
            title=second_name + ' Functions',
        )

    if events_only or not functions_only:

        if functions_only or not events_only:
            print()
            print()

        contract_abi_summary.print_contract_abi_events(
            diff['mutual'],
            title='Mutual Events',
        )

        print()
        print()
        contract_abi_summary.print_contract_abi_events(
            diff['first_only'],
            title=first_name + ' Events',
        )

        print()
        print()
        contract_abi_summary.print_contract_abi_events(
            diff['second_only'],
            title=second_name + ' Events',
        )



from __future__ import annotations

import typing

from ctc import evm
from ctc import spec


url_templates: typing.Mapping[str, str] = {
    'abi': 'https://{api_subdomain}/api?module=contract&action=getabi&address={address}&format=raw',
    'address': 'https://{hostname}/address/{address}',
    'address_erc20_transfers': 'https://{hostname}/tokentxns?a={address}',
    'address_internal_txs': 'https://{hostname}/txsinternal?ps=100&zero=false&a={address}&valid=all&m=advanced',
    'address_holdings': 'https://{hostname}/tokenholdings?a={address}',
    'block': 'https://{hostname}/block/{block}',
    'token': 'https://{hostname}/token/{token_address}',
    'token_holders': 'https://{hostname}/token/{token_address}#balances',
    'transaction': 'https://{hostname}/tx/{transaction_hash}',
    'transaction_logs': 'https://{hostname}/tx/{transaction_hash}#eventlog',
    'transaction_state_changes': 'https://{hostname}/tx/{transaction_hash}#statechange',
}


def get_hostname(network: spec.NetworkReference) -> str:
    block_explorer = evm.get_network_block_explorer(network)
    if block_explorer is None:
        raise Exception('hostname unknown for network: ' + str(network))
    return block_explorer


def get_api_subdomain(network: spec.NetworkReference) -> str:

    non_etherscan_block_explorers = [
        'gnosis',
        'harmony',
        'harmony_testnet',
    ]
    api_dash_subdomains = [
        'ropsten',
        'kovan',
        'rinkeby',
        'goerli',
        'optimism',
        'kovan-optimistic',
        'bnb_testnet',
    ]

    block_explorer = get_hostname(network)
    network_name = evm.get_network_name(network)
    if network_name in non_etherscan_block_explorers:
        raise NotImplementedError(
            'no abi source for network: ' + str(network_name)
        )
    elif network_name in api_dash_subdomains:
        return 'api-' + block_explorer
    else:
        return 'api.' + block_explorer


def create_abi_url(
    address: spec.Address, network: spec.NetworkReference = 'ethereum'
) -> str:
    api_subdomain = get_api_subdomain(network)
    template = url_templates['abi']
    return template.format(api_subdomain=api_subdomain, address=address)


def create_address_url(
    address: spec.Address,
    network: spec.NetworkReference = 'ethereum',
) -> str:
    hostname = get_hostname(network)
    template = url_templates['address']
    return template.format(hostname=hostname, address=address)


def create_address_erc20_transfers_url(
    address: spec.Address,
    network: spec.NetworkReference = 'ethereum',
) -> str:
    hostname = get_hostname(network)
    template = url_templates['address_erc20_transfers']
    return template.format(hostname=hostname, address=address)


def create_address_internal_txs_url(
    address: spec.Address,
    network: spec.NetworkReference = 'ethereum',
) -> str:
    hostname = get_hostname(network)
    template = url_templates['address_internal_txs']
    return template.format(hostname=hostname, address=address)


def create_address_holdings_url(
    address: spec.Address,
    network: spec.NetworkReference = 'ethereum',
) -> str:
    hostname = get_hostname(network)
    template = url_templates['address_holdings']
    return template.format(hostname=hostname, address=address)


def create_block_url(
    block: int,
    network: spec.NetworkReference = 'ethereum',
) -> str:
    hostname = get_hostname(network)
    template = url_templates['block']
    return template.format(hostname=hostname, block=block)


def create_token_url(
    token_address: spec.Address,
    network: spec.NetworkReference = 'ethereum',
) -> str:
    hostname = get_hostname(network)
    template = url_templates['token']
    return template.format(hostname=hostname, token_address=token_address)


def create_token_holders_url(
    token_address: spec.Address,
    network: spec.NetworkReference = 'ethereum',
) -> str:
    hostname = get_hostname(network)
    template = url_templates['token_holders']
    return template.format(hostname=hostname, token_address=token_address)


def create_transaction_url(
    transaction_hash: str,
    network: spec.NetworkReference = 'ethereum',
) -> str:
    hostname = get_hostname(network)
    template = url_templates['transaction']
    return template.format(hostname=hostname, transaction_hash=transaction_hash)


def create_transaction_logs_url(
    transaction_hash: str,
    network: spec.NetworkReference = 'ethereum',
) -> str:
    hostname = get_hostname(network)
    template = url_templates['transaction_logs']
    return template.format(hostname=hostname, transaction_hash=transaction_hash)


def create_transaction_state_changes_url(
    transaction_hash: str,
    network: spec.NetworkReference = 'ethereum',
) -> str:
    hostname = get_hostname(network)
    template = url_templates['transaction_state_changes']
    return template.format(hostname=hostname, transaction_hash=transaction_hash)





import toolsql
import pytest

import ctc
from ctc import db

import conftest


def test_initialize_schema_version_schema():
    db_config = conftest.get_test_db_config()
    with toolsql.connect(db_config) as conn:
        db.initialize_schema_versions(conn=conn)
    with toolsql.connect(db_config) as conn:
        assert db.is_schema_versions_initialized(conn=conn)


@pytest.mark.parametrize(
    'schema_name', db.get_evm_schema_names() + db.get_generic_schema_names()
)
def test_initialize_schemas(schema_name):

    db_config = conftest.get_test_db_config()
    network = 1

    # create schema
    with toolsql.connect(db_config) as conn:
        db.initialize_schema(
            schema_name=schema_name,
            network=network,
            conn=conn,
        )

    # assert that schema version being tracked
    with toolsql.connect(db_config) as conn:
        schema_version = db.get_schema_version(
            schema_name=schema_name,
            context=dict(network=network),
            conn=conn,
        )
        assert schema_version == ctc.__version__




from __future__ import annotations

import toolsql

from ctc import spec
from . import erc20_metadata_statements


async def async_intake_erc20_metadata(
    address: spec.Address,
    *,
    context: spec.Context,
    decimals: int | None = None,
    symbol: str | None = None,
    name: str | None = None,
) -> None:

    from ctc import config

    db_config = config.get_context_db_config(
        schema_name='erc20_metadata',
        context=context,
    )
    async with toolsql.async_connect(db_config) as conn:
        await erc20_metadata_statements.async_upsert_erc20_metadata(
            conn=conn,
            address=address,
            decimals=decimals,
            symbol=symbol,
            name=name,
            context=context,
        )



from __future__ import annotations

from ctc import rpc
from ctc import spec

from .. import rari_abis


async def async_get_ctoken_comptroller(
    ctoken: spec.Address,
    block: spec.BlockNumberReference = 'latest',
) -> spec.Address:
    result = await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['comptroller'],
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


async def async_get_ctoken_underlying(
    ctoken: spec.Address,
    block: spec.BlockNumberReference = 'latest',
) -> spec.Address:
    result = await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['underlying'],
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


async def async_get_ctoken_irm(
    ctoken: spec.Address,
    block: spec.BlockNumberReference = 'latest',
) -> spec.Address:
    result = await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['interestRateModel'],
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result



from __future__ import annotations

import toolcli

from ctc import evm
from ctc import spec
from ctc.protocols import ens_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_resolve_command,
        'help': 'resolve ENS name',
        'args': [
            {'name': 'name', 'help': 'ENS name'},
            {'name': '--block', 'help': 'block number'},
        ],
        'examples': [
            'vitalik.eth',
            'vitalik.eth --block 14000000',
        ],
    }


async def async_resolve_command(
    name: str, block: spec.BlockNumberReference
) -> None:
    if block is not None:
        block = evm.standardize_block_number(block)
    address = await ens_utils.async_resolve_name(name, block=block)
    if address is not None:
        print(address)
    else:
        print('[name does not resolve to address]')



from __future__ import annotations

import toolcli

from ctc import evm

def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_transaction_command,
        'help': 'display transaction data, receipt, call data, & logs',
        'args': [
            {'name': 'transaction', 'help': 'hash of transaction'},
            {'name': '--sort', 'help': 'attribute to sort logs by'},
            {
                'name': '--json',
                'help': 'output raw JSON of tx',
                'action': 'store_true',
                'dest': 'as_json',
            },
            {
                'name': '--receipt',
                'help': 'output transaction receipt (only works with --json)',
                'action': 'store_true',
                'dest': 'receipt',
            },
        ],
        'examples': [
            '0xe981fe5c78d11d935a1dc35c579969e65e2dd6bb05ad321ea9670f8b1e203eaf',
            '0xe981fe5c78d11d935a1dc35c579969e65e2dd6bb05ad321ea9670f8b1e203eaf --json',
            '0xe981fe5c78d11d935a1dc35c579969e65e2dd6bb05ad321ea9670f8b1e203eaf --json --receipt',
        ],
    }


async def async_transaction_command(
    *, transaction: str, sort: str, as_json: bool, receipt: bool
) -> None:

    if as_json:
        import json
        import asyncio
        from ctc import rpc

        if receipt:
            receipt_task = asyncio.create_task(
                rpc.async_eth_get_transaction_receipt(
                    transaction_hash=transaction
                )
            )
            receipt_data = await receipt_task
            print(json.dumps(receipt_data, sort_keys=True, indent=4))
            return

        transaction_data = await evm.async_get_transaction(transaction)
        print(json.dumps(transaction_data, sort_keys=True, indent=4))

    else:
        if receipt:
            print('receipt flag only works with --json')
            return
        
        await evm.async_print_transaction_summary(
            transaction_hash=transaction, sort_logs_by=sort
        )



import toolsql

from ctc import db

import conftest


example_data = [
    {
        'address': '0x956f47f50a910163d8bf957cf5846d573e7f87ca',
        'abi': {'fake': 'abi', 'nested': {'attribute': 'value'}},
        'includes_proxy': False,
    },
    {
        'address': '0x9928e4046d7c6513326ccea028cd3e7a91c7590a',
        'abi': {
            'another_fake': 'abi',
            'another_nested': {'attribute': 'value'},
        },
        'includes_proxy': False,
    },
]


async def test_contract_abis_crud():

    db_config = conftest.get_test_db_config()
    db_schema = db.get_prepared_schema(
        schema_name='contract_abis',
        context=dict(network='ethereum'),
    )
    toolsql.create_db(
        db_config=db_config,
        db_schema=db_schema,
        if_not_exists=True,
        confirm=True,
    )

    # insert data
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            await db.async_upsert_contract_abi(conn=conn, **datum)

    # get data individually
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            db_contract_abi = await db.async_select_contract_abi(
                conn=conn,
                address=datum['address'],
            )
            assert datum['abi'] == db_contract_abi
            # for key, target_value in datum.items():
            #     assert target_value == db_contract_abi[key]

    # get data collectively
    all_addresses = [datum['address'] for datum in example_data]
    async with toolsql.async_connect(db_config) as conn:
        db_contract_abis = await db.async_select_contract_abis(
            conn=conn,
            addresses=all_addresses,
        )
        packaged_example_data = {
            datum['address']: datum['abi'] for datum in example_data
        }
        assert db_contract_abis == packaged_example_data

    # delete entries one by one
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            await db.async_delete_contract_abi(
                conn=conn,
                address=datum['address'],
            )

    # ensure all entries deleted
    async with toolsql.async_connect(db_config) as conn:
        db_contract_abis = await db.async_select_contract_abis(
            conn=conn,
            addresses=all_addresses,
        )
        assert all(item is None for item in db_contract_abis)

    # insert data again
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            await db.async_upsert_contract_abi(conn=conn, **datum)




from __future__ import annotations

from ctc import evm
from ctc import spec
from . import abi_crud
from . import url_crud


async def async_open_etherscan_in_browser(
    query: str,
    *,
    logs: bool | None = None,
    changes: bool | None = None,
    holdings: bool | None = None,
    erc20_transfers: bool | None = None,
    internal: bool | None = None,
    holders: bool | None = None,
    abi: bool | None = None,
    context: spec.Context | None = None,
) -> None:

    import toolcli
    import ctc.config

    if ctc.config.get_context_chain_id(context) != 1:
        raise NotImplementedError('only currently implemented for ethereum')

    is_address = False
    if query.startswith('0x') and len(query) == 42:
        url = url_crud.create_address_url(query)
        address = query
        is_address = True
    elif query.startswith('0x') and len(query) == 66:
        if logs:
            url = url_crud.create_transaction_logs_url(query)
        elif changes:
            url = url_crud.create_transaction_state_changes_url(query)
        else:
            url = url_crud.create_transaction_url(query)
    elif query.isnumeric():
        url = url_crud.create_block_url(int(query))
    elif query.endswith('.eth'):
        from ctc.protocols import ens_utils

        resolved_address = await ens_utils.async_resolve_name(query)
        if resolved_address is None:
            print('[name does not resolve to an address]')
        else:
            address = resolved_address
            url = url_crud.create_address_url(address)
            is_address = True
    else:
        try:
            address = await evm.async_get_erc20_address(query)
            url = url_crud.create_address_url(address)
            is_address = True
        except Exception:
            raise Exception('unknown query: ' + str(query))

    if is_address:
        if holdings:
            url = url_crud.create_address_holdings_url(address)
        elif erc20_transfers:
            url = url_crud.create_address_erc20_transfers_url(address)
        elif internal:
            url = url_crud.create_address_internal_txs_url(address)
        elif holders:
            url = url_crud.create_token_holders_url(address)

    if abi:
        if is_address:
            import json

            raw_abi = await abi_crud.async_get_contract_abi(
                address, verbose=False
            )
            print(json.dumps(raw_abi, sort_keys=True, indent=4))
        else:
            raise Exception('can only use abi option for addresses')
    else:
        toolcli.open_url_in_browser(url)




from __future__ import annotations

import typing

from ctc import evm
from ctc import spec

from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


def sync_eth_call(
    to_address: spec.Address,
    *,
    from_address: spec.BinaryData | None = None,
    gas: spec.BinaryData | None = None,
    gas_price: spec.BinaryData | None = None,
    value_sent: spec.BinaryData | None = None,
    block_number: spec.BlockNumberReference | None = None,
    call_data: spec.BinaryData | None = None,
    function_parameters: typing.Sequence[typing.Any]
    | typing.Mapping[str, typing.Any]
    | None = None,
    context: spec.Context = None,
    decode_response: bool = True,
    delist_single_outputs: bool = True,
    package_named_outputs: bool = False,
    fill_empty: bool = False,
    empty_token: typing.Any = None,
    convert_reverts_to: typing.Any = None,
    convert_reverts_to_none: bool = False,
    function_abi: spec.FunctionABI | None = None,
    function_name: typing.Optional[str] = None,
    contract_abi: typing.Optional[spec.ContractABI] = None,
    n_parameters: typing.Optional[int] = None,
    parameter_types: typing.Optional[list[spec.ABIDatumType]] = None,
    function_selector: typing.Optional[spec.FunctionSelector] = None,
) -> spec.RpcSingularResponse:
    if function_abi is None:
        if call_data is None or decode_response:
            function_abi = evm.sync_get_function_abi(
                contract_address=to_address,
                contract_abi=contract_abi,
                function_name=function_name,
                n_parameters=n_parameters,
                parameter_types=parameter_types,
                function_selector=function_selector,
                context=context,
            )

    # construct request
    request = rpc_constructors.construct_eth_call(
        to_address=to_address,
        from_address=from_address,
        gas=gas,
        gas_price=gas_price,
        value_sent=value_sent,
        block_number=block_number,
        call_data=call_data,
        function_parameters=function_parameters,
        function_abi=function_abi,
    )

    # make request
    response = rpc_request.sync_send(
        request,
        context=context,
        convert_reverts_to_none=convert_reverts_to_none,
        convert_reverts_to=convert_reverts_to,
    )

    # digest response
    return rpc_digestors.digest_eth_call(
        response,
        function_abi=function_abi,
        decode_response=decode_response,
        delist_single_outputs=delist_single_outputs,
        package_named_outputs=package_named_outputs,
        fill_empty=fill_empty,
        empty_token=empty_token,
    )


def sync_eth_estimate_gas(
    to_address: spec.Address,
    *,
    from_address: spec.BinaryData | None = None,
    gas: spec.BinaryData | None = None,
    gas_price: spec.BinaryData | None = None,
    value_sent: spec.BinaryData | None = None,
    call_data: spec.BinaryData | None = None,
    function_parameters: typing.Sequence[typing.Any]
    | typing.Mapping[str, typing.Any]
    | None = None,
    context: spec.Context = None,
    decode_response: bool = True,
    function_abi: spec.FunctionABI | None = None,
    function_name: typing.Optional[str] = None,
    contract_abi: typing.Optional[spec.ContractABI] = None,
    n_parameters: typing.Optional[int] = None,
    parameter_types: typing.Optional[list[spec.ABIDatumType]] = None,
    function_selector: typing.Optional[spec.FunctionSelector] = None,
) -> spec.RpcSingularResponse:
    if function_abi is None:
        function_abi = evm.sync_get_function_abi(
            contract_address=to_address,
            contract_abi=contract_abi,
            function_name=function_name,
            n_parameters=n_parameters,
            parameter_types=parameter_types,
            function_selector=function_selector,
            context=context,
        )

    request = rpc_constructors.construct_eth_estimate_gas(
        to_address=to_address,
        from_address=from_address,
        gas=gas,
        gas_price=gas_price,
        value_sent=value_sent,
        call_data=call_data,
        function_parameters=function_parameters,
        function_abi=function_abi,
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_estimate_gas(
        response,
        decode_response=decode_response,
    )


def sync_eth_get_balance(
    address: spec.Address,
    *,
    block_number: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
    decode_response: bool = True,
) -> spec.RpcSingularResponse:
    if block_number is None:
        block_number = 'latest'
    request = rpc_constructors.construct_eth_get_balance(
        address=address,
        block_number=block_number,
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_get_balance(
        response,
        decode_response=decode_response,
    )


def sync_eth_get_storage_at(
    address: spec.Address,
    position: spec.BinaryData,
    *,
    block_number: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    if block_number is None:
        block_number = 'latest'
    request = rpc_constructors.construct_eth_get_storage_at(
        address=address,
        position=position,
        block_number=block_number,
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_get_storage_at(response)


def sync_eth_get_code(
    address: spec.Address,
    *,
    block_number: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    if block_number is None:
        block_number = 'latest'
    request = rpc_constructors.construct_eth_get_code(
        address=address,
        block_number=block_number,
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_get_code(response)




from __future__ import annotations

import time

import toolcli
import tooltime

from ctc import evm
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_block_command,
        'help': 'summarize block',
        'args': [
            {'name': 'block', 'nargs': '?', 'help': 'block number'},
            {
                'name': '--timestamp',
                'help': 'specify block by timestamp',
            },
            {
                'name': ['--verbose', '-v'],
                'action': 'store_true',
                'help': 'emit extra information',
            },
            {
                'name': '--json',
                'help': 'output block as raw json',
                'dest': 'as_json',
                'action': 'store_true',
            },
        ],
        'extra_data': ['parse_spec'],
        'examples': [
            '14000000',
            '14000000 --json',
            '--timestamp 1600000000',
        ],
    }


async def async_block_command(
    *,
    block: str | None,
    timestamp: str | int,
    verbose: bool,
    as_json: bool,
    parse_spec: toolcli.ParseSpec,
) -> None:

    if block is None:
        if timestamp is not None:

            if not isinstance(timestamp, str):
                raise Exception('function input should be str')

            if len(timestamp) == 4 or (
                len(timestamp) == 10 and timestamp.count('-') == 2
            ):
                pass
            else:
                timestamp = int(timestamp)

            timestamp_seconds = tooltime.timestamp_to_seconds(timestamp)

            if timestamp_seconds > time.time():
                block_number = await evm.async_predict_timestamp_block(
                    timestamp
                )
                print('predicting future block...')
                print(block_number)
                return
            else:
                print('searching for block at timestamp:', timestamp)
                block_number = await evm.async_get_block_of_timestamp(
                    timestamp=timestamp,
                    verbose=False,
                )
                print('found block', block_number)
                print()
        else:
            block_number = 'latest'
    else:
        block_number = await evm.async_block_number_to_int(block)

    if as_json:
        import rich
        import json

        block_data = await evm.async_get_block(block_number)
        rich.print_json(json.dumps(block_data))
    else:
        await evm.async_print_block_summary(block=block_number)



from __future__ import annotations

import typing

from ctc import evm
from ctc import rpc
from ctc import spec

from . import chainlink_db
from . import chainlink_spec


def _build_feed_query(
    feed: chainlink_spec._FeedReference,
) -> typing.Mapping[str, typing.Any]:
    if isinstance(feed, str):
        if evm.is_address_str(feed):
            return {'address': feed}
        elif '_' in feed:
            return {'name': feed.replace('_', ' / ')}
        else:
            return {'name': feed}
    else:
        raise Exception('unknown feed reference type')


async def async_get_feed_metadata(
    feed: chainlink_spec._FeedReference,
    *,
    context: spec.Context = None,
) -> chainlink_db.ChainlinkFeed:
    query = _build_feed_query(feed)
    feed_data = await chainlink_db.async_query_feed(context=context, **query)
    if feed_data is None:
        raise Exception('could not find data for feed: ' + str(feed))
    return feed_data


async def async_get_feed_decimals(
    feed: chainlink_spec._FeedReference,
    *,
    context: spec.Context = None,
) -> int:

    from ctc import config

    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='chainlink', context=context
    )

    # try database
    if read_cache:
        query = _build_feed_query(feed)
        feed_data = await chainlink_db.async_query_feed(
            context=context, **query
        )
        if feed_data is not None:
            return feed_data['decimals']

    # query rpc
    if evm.is_address_str(feed):
        result = await rpc.async_eth_call(
            feed,
            function_abi=chainlink_spec.feed_function_abis['decimals'],
            context=context,
        )
        if not isinstance(result, int):
            raise Exception('invalid rpc result')
        return result

    else:
        raise Exception('could not find address for feed: ' + str(feed))


async def async_resolve_feed_address(
    feed: str,
    *,
    context: spec.Context = None,
) -> spec.Address:

    if evm.is_address_str(feed):
        return feed

    query = _build_feed_query(feed)
    feed_data = await chainlink_db.async_query_feed(context=context, **query)
    if feed_data is not None:
        return feed_data['address']
    else:
        raise Exception('could not resolve feed address')


async def async_get_feed_aggregator(
    feed: chainlink_spec._FeedReference,
    *,
    block: spec.BlockNumberReference = 'latest',
    fill_empty: bool = True,
    context: spec.Context = None,
) -> spec.Address:

    feed = await async_resolve_feed_address(feed, context=context)

    aggregator = await rpc.async_eth_call(
        to_address=feed,
        function_abi=chainlink_spec.feed_function_abis['aggregator'],
        block_number=block,
        fill_empty=fill_empty,
        context=context,
    )
    if aggregator is None:
        raise Exception('aggregator not specified')
    if not isinstance(aggregator, str):
        raise Exception('invalid rpc result')

    return aggregator


async def async_get_feed_first_block(
    feed: chainlink_spec._FeedReference,
    *,
    start_search: typing.Optional[spec.BlockNumberReference] = None,
    end_search: typing.Optional[spec.BlockNumberReference] = None,
    verbose: bool = False,
    context: spec.Context = None,
) -> int:

    feed = await async_resolve_feed_address(feed, context=context)

    creation_block = await evm.async_get_contract_creation_block(
        contract_address=feed,
        start_block=start_search,
        end_block=end_search,
        verbose=verbose,
        context=context,
    )

    if creation_block is None:
        raise Exception('could not determine creation_block for feed')

    return creation_block




from __future__ import annotations

import typing

from ctc import spec

from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


async def async_shh_version(
    *,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_shh_version()
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_shh_version(
        response=response,
    )


async def async_shh_post(
    *,
    from_address: spec.BinaryData,
    to_address: spec.BinaryData,
    topics: list[spec.BinaryData],
    payload: spec.BinaryData,
    priority: spec.BinaryData,
    ttl: spec.BinaryData,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_shh_post(
        from_address=from_address,
        to_address=to_address,
        topics=topics,
        payload=payload,
        priority=priority,
        ttl=ttl,
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_shh_post(
        response=response,
    )


async def async_shh_new_identity(
    *,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_shh_new_identity()
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_shh_new_identity(response=response)


async def async_shh_has_identity(
    data: spec.BinaryData,
    *,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_shh_has_identity(data=data)
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_shh_has_identity(response=response)


async def async_shh_new_group(
    *,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_shh_new_group()
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_shh_new_group(response=response)


async def async_shh_add_to_group(
    data: spec.BinaryData, *, context: spec.Context = None
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_shh_add_to_group(data=data)
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_shh_add_to_group(response=response)


async def async_shh_new_filter(
    to_address: spec.Address,
    topics: typing.Sequence[spec.BinaryData],
    *,
    context: spec.Context = None,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_shh_new_filter(
        to_address=to_address, topics=topics
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_shh_new_filter(response=response)


async def async_shh_uninstall_filter(
    filter_id: spec.BinaryData, *, context: spec.Context = None
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_shh_uninstall_filter(
        filter_id=filter_id
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_shh_uninstall_filter(
        response=response,
    )


async def async_shh_get_filter_changes(
    filter_id: spec.BinaryData, *, context: spec.Context = None
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_shh_get_filter_changes(
        filter_id=filter_id,
    )
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_shh_get_filter_changes(
        response=response,
    )


async def async_shh_get_messages(
    filter_id: spec.BinaryData, *, context: spec.Context = None
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_shh_get_messages(filter_id=filter_id)
    response = await rpc_request.async_send(request, context=context)
    return rpc_digestors.digest_shh_get_messages(
        response=response,
    )




from __future__ import annotations

import toolcli
import toolstr

from ctc import cli
from ctc import config
from ctc import evm
from ctc import spec


help_message = """register custom proxy implementation of a contract

Implementation ABI will be downloaded for original contract

Manual registration using this command is normally unnecessary

This command is useful for custom proxy implementations"""


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_proxy_add_command,
        'help': help_message,
        'args': [
            {
                'name': 'proxy_contract',
                'help': 'contract that falls back to other contract',
            },
            {
                'name': 'implementation_contract',
                'help': 'contract that implements proxy contract',
            },
            {
                'name': '--confirm',
                'help': 'confirm registration of proxy',
                'action': 'store_true',
            },
        ],
        'examples': [
            '0xfb558ecd2d24886e8d2956775c619deb22f154ef 0xe16db319d9da7ce40b666dd2e365a4b8b3c18217 --confirm',
        ],
    }


async def async_proxy_add_command(
    proxy_contract: spec.Address,
    implementation_contract: spec.Address,
    *,
    confirm: bool,
) -> None:

    styles = cli.get_cli_styles()

    context: spec.Context = {}
    network = config.get_context_chain_id(context)

    if not confirm:
        toolstr.print(
            'Registering ABI implementation for use with proxy contract...',
            style=styles['description'],
        )
        toolstr.print(
            '             proxy '
            + toolstr.add_style(proxy_contract, styles['metavar']),
            style=styles['option'],
        )
        toolstr.print(
            '    implementation '
            + toolstr.add_style(implementation_contract, styles['metavar']),
            style=styles['option'],
        )
        toolstr.print(
            '           network '
            + toolstr.add_style(str(network), styles['metavar']),
            style=styles['option'],
        )
        print()
        answer = toolcli.input_yes_or_no(
            'Proceed? (y/n) ', style=styles['metavar']
        )
        if not answer:
            return

    context = config.create_user_input_context(network=network, cache=False)
    abi = await evm.async_get_contract_abi(
        contract_address=proxy_contract,
        proxy_implementation=implementation_contract,
        context=context,
    )

    print('resulting abi contains', len(abi), 'items')



from __future__ import annotations

import typing

from ctc import spec


async def async_compute_block_eth_prices(
    blocks: typing.Sequence[int] | spec.Series,
    context: spec.Context = None,
) -> spec.Series:
    import polars as pl
    from ctc.protocols import chainlink_utils

    start_block = min(blocks)
    end_block = max(blocks)
    ETH_USD = await chainlink_utils.async_get_feed_data(
        'ETH_USD',
        context=context,
        interpolate=True,
        start_block=start_block,
        end_block=end_block,
    )
    df = pl.DataFrame(blocks.alias('block_number'))
    joined = df.join(ETH_USD, on='block_number', how='left')
    return joined['answer'].alias('ETH_USD')




from __future__ import annotations

import typing

from ctc import spec
from .. import fourbyte_db
from .. import fourbyte_spec
from . import local_queries
from . import remote_queries


async def async_query_function_signatures(
    hex_signature: typing.Optional[str] = None,
    *,
    id: typing.Optional[int] = None,
    # bytes_signature: typing.Optional[str] = None,
    text_signature: typing.Optional[str] = None,
    use_local: bool = True,
    use_remote: bool = True,
    context: spec.Context = None,
) -> typing.Sequence[fourbyte_spec.Entry]:
    if not use_local and not use_remote:
        raise Exception('should use at least one of use_local or use_remote')

    if use_local:
        result = await local_queries.async_query_local_function_signatures(
            id=id,
            # bytes_signature=bytes_signature,
            hex_signature=hex_signature,
            text_signature=text_signature,
            context={},
        )
        if result is not None and len(result) > 0:
            return result

    if use_remote:
        result = await remote_queries.async_query_remote_function_signatures(
            id=id,
            # bytes_signature=bytes_signature,
            hex_signature=hex_signature,
            text_signature=text_signature,
        )

        # type ignoring because of mypy bug
        await fourbyte_db.async_intake_function_signatures(result)  # type: ignore

        return result

    return []


async def async_query_event_signatures(
    hex_signature: typing.Optional[str] = None,
    *,
    id: typing.Optional[int] = None,
    # bytes_signature: typing.Optional[str] = None,
    text_signature: typing.Optional[str] = None,
    use_local: bool = True,
    use_remote: bool = True,
) -> typing.Sequence[fourbyte_spec.Entry]:
    if not use_local and not use_remote:
        raise Exception('should use at least one of use_local or use_remote')

    if use_local:
        result = await local_queries.async_query_local_event_signatures(
            id=id,
            # bytes_signature=bytes_signature,
            hex_signature=hex_signature,
            text_signature=text_signature,
            context={},
        )
        if result is not None and len(result) > 0:
            return result

    if use_remote:
        result = await remote_queries.async_query_remote_event_signatures(
            id=id,
            # bytes_signature=bytes_signature,
            hex_signature=hex_signature,
            text_signature=text_signature,
        )

        # type ignoring because of mypy bug
        await fourbyte_db.async_intake_event_signatures(
            typing.cast(typing.Sequence[fourbyte_spec.PartialEntry], result)
        )

        return result

    return []




import pytest

from ctc import evm


def test_pack_vrs_transaction():
    # from https://ethereum.stackexchange.com/a/107498
    target_signature = '0x28a0da4429a9e8e6b54cb101b2df002039f2879ab4ca0e8fae64134942cb81f3e581a03b90a37dc078a82dfc418695b1d4473661aa4d24dd874ac68678894ff44a6b27'
    actual_signature = evm.pack_signature_vrs(
        v='0x28',
        r='0xda4429a9e8e6b54cb101b2df002039f2879ab4ca0e8fae64134942cb81f3e581',
        s='0x3b90a37dc078a82dfc418695b1d4473661aa4d24dd874ac68678894ff44a6b27',
        mode='transaction',
    )
    assert target_signature == actual_signature


private_key_to_public_key_examples = [
    [
        '0x080a12470a639f95139e5e2d9fc7ca597869a42de9bfab4969a3a57a89b0c84a',
        (
            28861520271981833226067272609620060740909575585894736662694419241176299067921,
            104228060715000826995481435060948770188121251882107489647159844236681839901412,
        ),
    ],
    [
        # from https://github.com/ethereum/py_ecc/blob/033b4ea69a3cbc841c02c9b7d1b4996cd7863585/tests/test_secp256k1.py
        '792eca682b890b31356247f2b04662bff448b6bb19ea1c8ab48da222c894ef9b',
        (
            20033694065814990006010338153307081985267967222430278129327181081381512401190,
            72089573118161052907088366229362685603474623289048716349537937839432544970413,
        ),
    ],
]


@pytest.mark.parametrize('test', private_key_to_public_key_examples)
def test_private_key_to_public_key(test):
    private_key, target_public_key = test
    actual_public_key = evm.private_key_to_public_key(private_key)
    assert (
        evm.public_key_tuple_to_hex(target_public_key) == actual_public_key
    )


private_key_to_address_examples = [
    [
        '0x080a12470a639f95139e5e2d9fc7ca597869a42de9bfab4969a3a57a89b0c84a',
        '0x774246187e1e2205c5920898eede0945016080df',
    ],
    [
        b"\xb2\\}\xb3\x1f\xee\xd9\x12''\xbf\t9\xdcv\x9a\x96VK-\xe4\xc4rm\x03[6\xec\xf1\xe5\xb3d",
        '0x5ce9454909639d2d17a3f753ce7d93fa0b9ab12e',
    ],
]


@pytest.mark.parametrize('test', private_key_to_address_examples)
def test_private_key_to_address(test):
    private_key, target_address = test
    actual_address = evm.private_key_to_address(private_key)
    assert target_address == actual_address


sign_text_message_examples = [
    {
        # from https://github.com/ethereum/web3.py/blob/master/docs/web3.eth.account.rst#sign-a-message
        'message': "I♥SF",
        'private_key': b"\xb2\\}\xb3\x1f\xee\xd9\x12''\xbf\t9\xdcv\x9a\x96VK-\xe4\xc4rm\x03[6\xec\xf1\xe5\xb3d",
        'target_signature': (
            28,
            104389933075820307925104709181714897380569894203213074526835978196648170704563,
            28205917190874851400050446352651915501321657673772411533993420917949420456142,
        ),
        'mode': 'personal_sign',
    },
]


@pytest.mark.parametrize('test', sign_text_message_examples)
def test_sign_text_message(test):
    message = test['message']
    private_key = test['private_key']
    target_signature = test['target_signature']
    mode = test['mode']

    actual_signature = evm.sign_text_message(
        message=message,
        private_key=private_key,
        mode=mode,
    )
    assert target_signature == actual_signature


sign_message_hash_examples = [
    {
        # from EIP-155 https://eips.ethereum.org/EIPS/eip-155
        'message_hash': '0xdaf5a779ae972f972197303d7b574746c7ef83eadac0f2791ad23db92e4c8e53',
        'private_key': '0x4646464646464646464646464646464646464646464646464646464646464646',
        'target_signature': (
            37,
            18515461264373351373200002665853028612451056578545711640558177340181847433846,
            46948507304638947509940763649030358759909902576025900602547168820602576006531,
        ),
    },
]


@pytest.mark.parametrize('test', sign_message_hash_examples)
def test_sign_message_hash(test):
    message_hash = test['message_hash']
    private_key = test['private_key']
    target_signature = test['target_signature']

    actual_signature = evm.sign_message_hash(
        message_hash=message_hash,
        private_key=private_key,
        chain_id=1,
    )

    assert target_signature == actual_signature


@pytest.mark.parametrize('test', sign_message_hash_examples)
def test_recover_address(test):
    test = sign_message_hash_examples[0]

    signature = evm.sign_message_hash(
        message_hash=test['message_hash'],
        private_key=test['private_key'],
        chain_id=1,
    )

    target_public_key = evm.private_key_to_public_key(test['private_key'])
    actual_public_key = evm.recover_signer_public_key(
        message_hash=test['message_hash'],
        signature=signature,
    )
    assert actual_public_key == target_public_key

    assert evm.verify_signature(
        signature=signature,
        message_hash=test['message_hash'],
        public_key=target_public_key,
    )



from __future__ import annotations

import toolstr

import ctc.config
from ctc import cli
from . import llama_requests


#
# # comparisons
#


async def async_print_protocols_tvls(
    *,
    verbose: bool = False,
    n: int = 50,
    filter_category: str | None = None,
    filter_chain: str | None = None,
) -> None:

    styles = cli.get_cli_styles()

    data = await llama_requests.async_get_protocols_tvls(
        category=filter_category,
        chain=filter_chain,
    )

    keys = [
        'name',
        'category',
        'chain',
        'tvl',
    ]

    if verbose:
        keys.append('slug')

    rows = []
    for datum in data[:n]:
        name = ' '.join(datum['name'].split(' ')[:2])
        row = [name, datum['category'], datum['chain'], datum['tvl']]
        if verbose:
            row.append(datum['slug'])
        rows.append(row)
    if n < len(data):
        rows.append(['...'] * len(keys))

    toolstr.print_text_box('TVL by Protocol', style=styles['title'])
    print()
    toolstr.print_table(
        rows,
        labels=keys,
        column_formats={
            'tvl': {
                'order_of_magnitude': True,
                'decimals': 2,
                'trailing_zeros': True,
                'prefix': '$',
            },
        },
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'tvl': styles['description'],
        },
    )


async def async_print_chains_tvls(n: int = 50) -> None:

    styles = cli.get_cli_styles()

    data = await llama_requests.async_get_chains_tvls()

    keys = [
        'name',
        'tvl',
    ]

    data = sorted(data, key=lambda datum: float(datum['tvl']), reverse=True)

    rows = []
    for datum in data[:n]:
        name = ' '.join(datum['name'].split(' ')[:2])
        row = [name, datum['tvl']]
        rows.append(row)

    if len(data) > n:
        rows.append(['...'] * len(keys))

    toolstr.print_text_box('TVL by Chain', style=styles['title'])
    print()
    toolstr.print_table(
        rows,
        labels=keys,
        column_formats={
            'tvl': {
                'order_of_magnitude': True,
                'decimals': 2,
                'trailing_zeros': True,
                'prefix': '$',
            },
        },
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'tvl': styles['description'],
        },
    )


#
# # historical charting
#


async def async_print_historical_defi_tvl() -> None:

    styles = cli.get_cli_styles()

    data = await llama_requests.async_get_historical_defi_tvl()

    plot = toolstr.render_line_plot(
        xvals=data['timestamp'],
        yvals=data['tvl'],
        n_rows=10,
        n_columns=60,
        line_style=styles['description'],
        chrome_style=styles['comment'],
        tick_label_style=styles['metavar'],
        yaxis_kwargs={'tick_label_format': {'prefix': '$'}},
        char_dict=ctc.config.get_cli_chart_charset(),
    )

    toolstr.print_text_box('Historical Defi TVL', style=styles['title'])
    print()
    toolstr.print(plot)


async def async_print_historical_chain_tvl(chain: str) -> None:

    styles = cli.get_cli_styles()

    data = await llama_requests.async_get_historical_chain_tvl(chain)

    plot = toolstr.render_line_plot(
        xvals=data['timestamp'],
        yvals=data['tvl'],
        n_rows=10,
        n_columns=60,
        line_style=styles['description'],
        chrome_style=styles['comment'],
        tick_label_style=styles['metavar'],
        yaxis_kwargs={'tick_label_format': {'prefix': '$'}},
        char_dict=ctc.config.get_cli_chart_charset(),
    )

    toolstr.print_text_box(
        'Historical ' + chain + ' TVL', style=styles['title']
    )
    print()
    toolstr.print(plot, indent=4)


async def async_print_historical_protocol_tvl(
    protocol: str, verbose: bool = False
) -> None:

    styles = cli.get_cli_styles()

    uniswap = await llama_requests.async_get_historical_protocol_tvl(protocol)

    timestamp = []
    tvl = []
    for datum in uniswap['tvl']:
        timestamp.append(datum['date'])
        tvl.append(datum['totalLiquidityUSD'])

    plot = toolstr.render_line_plot(
        xvals=timestamp,
        yvals=tvl,
        n_rows=10,
        n_columns=60,
        line_style=styles['description'],
        chrome_style=styles['comment'],
        tick_label_style=styles['metavar'],
        yaxis_kwargs={'tick_label_format': {'prefix': '$'}},
        char_dict=ctc.config.get_cli_chart_charset(),
    )

    toolstr.print_text_box(
        'Historical ' + protocol + ' TVL', style=styles['title']
    )
    print()

    if verbose:
        toolstr.print(
            toolstr.hjustify('All Chains', 'center', 70),
            indent=4,
            style=styles['title'],
        )
    toolstr.print(plot, indent=4)

    if verbose:
        chains = sorted(
            uniswap['currentChainTvls'].keys(),
            key=lambda chain: float(uniswap['currentChainTvls'][chain]),
            reverse=True,
        )

        print()
        for chain in chains:
            raw_data = uniswap['chainTvls'][chain]['tvl']
            timestamp = []
            tvl = []
            for datum in raw_data:
                timestamp.append(datum['date'])
                tvl.append(datum['totalLiquidityUSD'])

            plot = toolstr.render_line_plot(
                xvals=timestamp,
                yvals=tvl,
                n_rows=10,
                n_columns=60,
                line_style=styles['description'],
                chrome_style=styles['comment'],
                tick_label_style=styles['metavar'],
                yaxis_kwargs={'tick_label_format': {'prefix': '$'}},
                char_dict=ctc.config.get_cli_chart_charset(),
            )
            print()
            toolstr.print(
                toolstr.hjustify(chain, 'center', 70),
                indent=4,
                style=styles['title'],
            )
            print()
            toolstr.print(
                plot,
                indent=4,
            )
            print()



from __future__ import annotations

import typing

import toolcli
import toolstr

from ctc import cli
from ctc import spec
from ctc.config import config_defaults


def setup_cli(
    *,
    styles: toolcli.StyleTheme,
    old_config: typing.Mapping[typing.Any, typing.Any],
    headless: bool,
) -> spec.PartialConfig:

    print()
    print()
    toolstr.print('## CLI Customization', style=styles['title'])

    print()
    cli_color_theme = _get_cli_color_theme(
        styles=styles,
        old_config=old_config,
        headless=headless,
    )
    print()
    cli_chart_charset = _get_cli_chart_charset(
        styles=styles,
        old_config=old_config,
        headless=headless,
    )

    return {
        'cli_color_theme': cli_color_theme,
        'cli_chart_charset': cli_chart_charset,
    }


def _get_cli_color_theme(
    *,
    styles: toolcli.StyleTheme,
    old_config: typing.Mapping[typing.Any, typing.Any],
    headless: bool,
) -> toolcli.StyleTheme:

    old_color_theme = old_config.get('cli_color_theme')
    if old_color_theme is None:
        old_color_theme = config_defaults.get_default_cli_color_theme()

    print('Current color theme:')
    for key, value in old_color_theme.items():
        cli.print_bullet(
            key=key,
            value=toolstr.add_style(str(value), str(value)),
            key_style='',
        )
    print()
    answer = toolcli.input_yes_or_no(
        'Modify this theme? ',
        default='no',
        style=styles['metavar'],
        headless=headless,
    )
    if answer:
        from ctc.cli import cli_utils

        current_theme_name = cli_utils.get_color_theme_name()
        using_custom_theme = current_theme_name not in cli_utils.color_themes

        print()
        print('Available color themes:')
        color_keys = [
            'title',
            'metavar',
            'description',
            'content',
            'option',
            'comment',
        ]
        cells = []
        if using_custom_theme:
            cell = toolstr.add_style('current', 'bold') + '\n'
            for color_key in color_keys:
                cell += '\n' + toolstr.add_style(
                    color_key, old_color_theme[color_key]  # type: ignore
                )
            cells.append(cell)
        for theme_name, theme_colors in cli_utils.color_themes.items():
            cell = toolstr.add_style(theme_name, 'bold') + '\n'
            for color_key in color_keys:
                cell += '\n' + toolstr.add_style(
                    color_key, theme_colors[color_key]  # type: ignore
                )
            cells.append(cell)
        cells_per_row = 6
        n_rows = _ceil_division(len(cells), cells_per_row)
        rows = [
            cells[r * cells_per_row: (r + 1) * cells_per_row]
            for r in range(n_rows)
        ]
        print()
        toolstr.print_multiline_table(rows)

        choices = list(cli_utils.color_themes.keys())
        if using_custom_theme:
            choices = ['current'] + choices
        choices.append('create custom theme')
        print()
        choice = toolcli.input_number_choice(
            prompt='Which color theme to use? ',
            choices=choices,
            default=choices[0],
        )
        choice_text = choices[choice]
        if choice_text in cli_utils.color_themes:
            new_color_theme = cli_utils.color_themes[choice_text]
        elif choice_text == 'create custom theme':
            new_color_theme = cli_utils.create_custom_color_theme(
                old_color_theme=cli.get_cli_styles(),
                headless=headless,
            )
        else:
            raise Exception('choice error')
        return new_color_theme
    else:
        return old_color_theme


def _ceil_division(a: int, b: int) -> int:
    # from https://stackoverflow.com/a/17511341
    # better than math.ceil because no floating point error
    return -(a // -b)


def _get_cli_chart_charset(
    *,
    styles: toolcli.StyleTheme,
    old_config: typing.Mapping[typing.Any, typing.Any],
    headless: bool,
) -> toolstr.SampleMode:
    return config_defaults.get_default_cli_chart_charset()




from __future__ import annotations

import toolcli

from .. import llama_tvls


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_llama_protocols_command,
        'help': 'output data about protocols tracked by Defi Llama',
        'args': [
            {
                'name': ['-v', '--verbose'],
                'help': 'show additional information',
                'action': 'store_true',
            },
            {
                'name': '--category',
                'help': 'filter protocols by category',
            },
            {
                'help': 'filter protocols by chain',
                'name': '--chain',
            },
            {
                'name': '-n',
                'help': 'number of protocols to display',
                'type': int,
                'default': 15,
            },
        ],
        'examples': [
            '',
            '--category Dexes',
            '--chain Tron',
        ],
    }


async def async_llama_protocols_command(
    *,
    verbose: bool = False,
    n: int = 50,
    category: str | None = None,
    chain: str | None = None,
) -> None:

    await llama_tvls.async_print_protocols_tvls(
        verbose=verbose,
        n=n,
        filter_category=category,
        filter_chain=chain,
    )



from __future__ import annotations

import typing

import toolcli

from ctc import evm
from ctc.cli import cli_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_transfers_command,
        'help': 'output information about ERC20 transfers',
        'args': [
            {'name': 'erc20', 'help': 'ERC20 address'},
            {
                'name': '--blocks',
                'help': 'block range of transfers',
            },
            {
                'name': '--include-timestamps',
                'default': False,
                'action': 'store_true',
                'help': 'include timestamps',
            },
            {
                'name': '--export',
                'default': 'stdout',
                'help': 'file path for output (.json or .csv)',
            },
            {
                'name': '--overwrite',
                'action': 'store_true',
                'help': 'specify that output path can be overwritten',
            },
        ],
        'examples': {
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca': {'long': True},
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca --blocks 14000000:14001000': {},
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca --blocks 14000000:14001000 --include-timestamps': {},
        },
    }


async def async_transfers_command(
    *,
    erc20: str,
    blocks: str | None,
    include_timestamps: bool,
    export: str,
    overwrite: bool,
) -> None:

    if blocks is not None:
        start_block, end_block = await cli_utils.async_parse_block_range(
            blocks
        )
    else:
        start_block = None
        end_block = None

    transfers = await evm.async_get_erc20_transfers(
        erc20,
        start_block=start_block,
        end_block=end_block,
        include_timestamps=include_timestamps,
    )

    cli_utils.output_data(transfers, output=export, overwrite=overwrite)



"""
TODO: move this data to directory
"""
from __future__ import annotations

import functools
import json
import typing

from ctc import evm
from ctc import spec


raw_primary_abi = """[{"inputs":[],"name":"directory","outputs":[{"internalType":"contract FusePoolDirectory","name":"","type":"address"}],"stateMutability":"view","type":"function","constant":true},{"inputs":[{"internalType":"contract FusePoolDirectory","name":"_directory","type":"address"}],"name":"initialize","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[],"name":"getPublicPoolsWithData","outputs":[{"internalType":"uint256[]","name":"","type":"uint256[]"},{"components":[{"internalType":"string","name":"name","type":"string"},{"internalType":"address","name":"creator","type":"address"},{"internalType":"address","name":"comptroller","type":"address"},{"internalType":"uint256","name":"blockPosted","type":"uint256"},{"internalType":"uint256","name":"timestampPosted","type":"uint256"}],"internalType":"struct FusePoolDirectory.FusePool[]","name":"","type":"tuple[]"},{"components":[{"internalType":"uint256","name":"totalSupply","type":"uint256"},{"internalType":"uint256","name":"totalBorrow","type":"uint256"},{"internalType":"address[]","name":"underlyingTokens","type":"address[]"},{"internalType":"string[]","name":"underlyingSymbols","type":"string[]"},{"internalType":"bool","name":"whitelistedAdmin","type":"bool"}],"internalType":"struct FusePoolLens.FusePoolData[]","name":"","type":"tuple[]"},{"internalType":"bool[]","name":"","type":"bool[]"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bool","name":"whitelistedAdmin","type":"bool"}],"name":"getPublicPoolsByVerificationWithData","outputs":[{"internalType":"uint256[]","name":"","type":"uint256[]"},{"components":[{"internalType":"string","name":"name","type":"string"},{"internalType":"address","name":"creator","type":"address"},{"internalType":"address","name":"comptroller","type":"address"},{"internalType":"uint256","name":"blockPosted","type":"uint256"},{"internalType":"uint256","name":"timestampPosted","type":"uint256"}],"internalType":"struct FusePoolDirectory.FusePool[]","name":"","type":"tuple[]"},{"components":[{"internalType":"uint256","name":"totalSupply","type":"uint256"},{"internalType":"uint256","name":"totalBorrow","type":"uint256"},{"internalType":"address[]","name":"underlyingTokens","type":"address[]"},{"internalType":"string[]","name":"underlyingSymbols","type":"string[]"},{"internalType":"bool","name":"whitelistedAdmin","type":"bool"}],"internalType":"struct FusePoolLens.FusePoolData[]","name":"","type":"tuple[]"},{"internalType":"bool[]","name":"","type":"bool[]"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"account","type":"address"}],"name":"getPoolsByAccountWithData","outputs":[{"internalType":"uint256[]","name":"","type":"uint256[]"},{"components":[{"internalType":"string","name":"name","type":"string"},{"internalType":"address","name":"creator","type":"address"},{"internalType":"address","name":"comptroller","type":"address"},{"internalType":"uint256","name":"blockPosted","type":"uint256"},{"internalType":"uint256","name":"timestampPosted","type":"uint256"}],"internalType":"struct FusePoolDirectory.FusePool[]","name":"","type":"tuple[]"},{"components":[{"internalType":"uint256","name":"totalSupply","type":"uint256"},{"internalType":"uint256","name":"totalBorrow","type":"uint256"},{"internalType":"address[]","name":"underlyingTokens","type":"address[]"},{"internalType":"string[]","name":"underlyingSymbols","type":"string[]"},{"internalType":"bool","name":"whitelistedAdmin","type":"bool"}],"internalType":"struct FusePoolLens.FusePoolData[]","name":"","type":"tuple[]"},{"internalType":"bool[]","name":"","type":"bool[]"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"contract Comptroller","name":"comptroller","type":"address"}],"name":"getPoolSummary","outputs":[{"internalType":"uint256","name":"","type":"uint256"},{"internalType":"uint256","name":"","type":"uint256"},{"internalType":"address[]","name":"","type":"address[]"},{"internalType":"string[]","name":"","type":"string[]"},{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"contract Comptroller","name":"comptroller","type":"address"}],"name":"getPoolAssetsWithData","outputs":[{"components":[{"internalType":"address","name":"cToken","type":"address"},{"internalType":"address","name":"underlyingToken","type":"address"},{"internalType":"string","name":"underlyingName","type":"string"},{"internalType":"string","name":"underlyingSymbol","type":"string"},{"internalType":"uint256","name":"underlyingDecimals","type":"uint256"},{"internalType":"uint256","name":"underlyingBalance","type":"uint256"},{"internalType":"uint256","name":"supplyRatePerBlock","type":"uint256"},{"internalType":"uint256","name":"borrowRatePerBlock","type":"uint256"},{"internalType":"uint256","name":"totalSupply","type":"uint256"},{"internalType":"uint256","name":"totalBorrow","type":"uint256"},{"internalType":"uint256","name":"supplyBalance","type":"uint256"},{"internalType":"uint256","name":"borrowBalance","type":"uint256"},{"internalType":"uint256","name":"liquidity","type":"uint256"},{"internalType":"bool","name":"membership","type":"bool"},{"internalType":"uint256","name":"exchangeRate","type":"uint256"},{"internalType":"uint256","name":"underlyingPrice","type":"uint256"},{"internalType":"address","name":"oracle","type":"address"},{"internalType":"uint256","name":"collateralFactor","type":"uint256"},{"internalType":"uint256","name":"reserveFactor","type":"uint256"},{"internalType":"uint256","name":"adminFee","type":"uint256"},{"internalType":"uint256","name":"fuseFee","type":"uint256"},{"internalType":"bool","name":"borrowGuardianPaused","type":"bool"}],"internalType":"struct FusePoolLens.FusePoolAsset[]","name":"","type":"tuple[]"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"uint256","name":"maxHealth","type":"uint256"}],"name":"getPublicPoolUsersWithData","outputs":[{"internalType":"contract Comptroller[]","name":"","type":"address[]"},{"components":[{"internalType":"address","name":"account","type":"address"},{"internalType":"uint256","name":"totalBorrow","type":"uint256"},{"internalType":"uint256","name":"totalCollateral","type":"uint256"},{"internalType":"uint256","name":"health","type":"uint256"},{"components":[{"internalType":"address","name":"cToken","type":"address"},{"internalType":"address","name":"underlyingToken","type":"address"},{"internalType":"string","name":"underlyingName","type":"string"},{"internalType":"string","name":"underlyingSymbol","type":"string"},{"internalType":"uint256","name":"underlyingDecimals","type":"uint256"},{"internalType":"uint256","name":"underlyingBalance","type":"uint256"},{"internalType":"uint256","name":"supplyRatePerBlock","type":"uint256"},{"internalType":"uint256","name":"borrowRatePerBlock","type":"uint256"},{"internalType":"uint256","name":"totalSupply","type":"uint256"},{"internalType":"uint256","name":"totalBorrow","type":"uint256"},{"internalType":"uint256","name":"supplyBalance","type":"uint256"},{"internalType":"uint256","name":"borrowBalance","type":"uint256"},{"internalType":"uint256","name":"liquidity","type":"uint256"},{"internalType":"bool","name":"membership","type":"bool"},{"internalType":"uint256","name":"exchangeRate","type":"uint256"},{"internalType":"uint256","name":"underlyingPrice","type":"uint256"},{"internalType":"address","name":"oracle","type":"address"},{"internalType":"uint256","name":"collateralFactor","type":"uint256"},{"internalType":"uint256","name":"reserveFactor","type":"uint256"},{"internalType":"uint256","name":"adminFee","type":"uint256"},{"internalType":"uint256","name":"fuseFee","type":"uint256"},{"internalType":"bool","name":"borrowGuardianPaused","type":"bool"}],"internalType":"struct FusePoolLens.FusePoolAsset[]","name":"assets","type":"tuple[]"}],"internalType":"struct FusePoolLens.FusePoolUser[][]","name":"","type":"tuple[][]"},{"internalType":"uint256[]","name":"","type":"uint256[]"},{"internalType":"uint256[]","name":"","type":"uint256[]"},{"internalType":"bool[]","name":"","type":"bool[]"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"contract Comptroller[]","name":"comptrollers","type":"address[]"},{"internalType":"uint256","name":"maxHealth","type":"uint256"}],"name":"getPoolUsersWithData","outputs":[{"components":[{"internalType":"address","name":"account","type":"address"},{"internalType":"uint256","name":"totalBorrow","type":"uint256"},{"internalType":"uint256","name":"totalCollateral","type":"uint256"},{"internalType":"uint256","name":"health","type":"uint256"},{"components":[{"internalType":"address","name":"cToken","type":"address"},{"internalType":"address","name":"underlyingToken","type":"address"},{"internalType":"string","name":"underlyingName","type":"string"},{"internalType":"string","name":"underlyingSymbol","type":"string"},{"internalType":"uint256","name":"underlyingDecimals","type":"uint256"},{"internalType":"uint256","name":"underlyingBalance","type":"uint256"},{"internalType":"uint256","name":"supplyRatePerBlock","type":"uint256"},{"internalType":"uint256","name":"borrowRatePerBlock","type":"uint256"},{"internalType":"uint256","name":"totalSupply","type":"uint256"},{"internalType":"uint256","name":"totalBorrow","type":"uint256"},{"internalType":"uint256","name":"supplyBalance","type":"uint256"},{"internalType":"uint256","name":"borrowBalance","type":"uint256"},{"internalType":"uint256","name":"liquidity","type":"uint256"},{"internalType":"bool","name":"membership","type":"bool"},{"internalType":"uint256","name":"exchangeRate","type":"uint256"},{"internalType":"uint256","name":"underlyingPrice","type":"uint256"},{"internalType":"address","name":"oracle","type":"address"},{"internalType":"uint256","name":"collateralFactor","type":"uint256"},{"internalType":"uint256","name":"reserveFactor","type":"uint256"},{"internalType":"uint256","name":"adminFee","type":"uint256"},{"internalType":"uint256","name":"fuseFee","type":"uint256"},{"internalType":"bool","name":"borrowGuardianPaused","type":"bool"}],"internalType":"struct FusePoolLens.FusePoolAsset[]","name":"assets","type":"tuple[]"}],"internalType":"struct FusePoolLens.FusePoolUser[][]","name":"","type":"tuple[][]"},{"internalType":"uint256[]","name":"","type":"uint256[]"},{"internalType":"uint256[]","name":"","type":"uint256[]"},{"internalType":"bool[]","name":"","type":"bool[]"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"contract Comptroller","name":"comptroller","type":"address"},{"internalType":"uint256","name":"maxHealth","type":"uint256"}],"name":"getPoolUsersWithData","outputs":[{"components":[{"internalType":"address","name":"account","type":"address"},{"internalType":"uint256","name":"totalBorrow","type":"uint256"},{"internalType":"uint256","name":"totalCollateral","type":"uint256"},{"internalType":"uint256","name":"health","type":"uint256"},{"components":[{"internalType":"address","name":"cToken","type":"address"},{"internalType":"address","name":"underlyingToken","type":"address"},{"internalType":"string","name":"underlyingName","type":"string"},{"internalType":"string","name":"underlyingSymbol","type":"string"},{"internalType":"uint256","name":"underlyingDecimals","type":"uint256"},{"internalType":"uint256","name":"underlyingBalance","type":"uint256"},{"internalType":"uint256","name":"supplyRatePerBlock","type":"uint256"},{"internalType":"uint256","name":"borrowRatePerBlock","type":"uint256"},{"internalType":"uint256","name":"totalSupply","type":"uint256"},{"internalType":"uint256","name":"totalBorrow","type":"uint256"},{"internalType":"uint256","name":"supplyBalance","type":"uint256"},{"internalType":"uint256","name":"borrowBalance","type":"uint256"},{"internalType":"uint256","name":"liquidity","type":"uint256"},{"internalType":"bool","name":"membership","type":"bool"},{"internalType":"uint256","name":"exchangeRate","type":"uint256"},{"internalType":"uint256","name":"underlyingPrice","type":"uint256"},{"internalType":"address","name":"oracle","type":"address"},{"internalType":"uint256","name":"collateralFactor","type":"uint256"},{"internalType":"uint256","name":"reserveFactor","type":"uint256"},{"internalType":"uint256","name":"adminFee","type":"uint256"},{"internalType":"uint256","name":"fuseFee","type":"uint256"},{"internalType":"bool","name":"borrowGuardianPaused","type":"bool"}],"internalType":"struct FusePoolLens.FusePoolAsset[]","name":"assets","type":"tuple[]"}],"internalType":"struct FusePoolLens.FusePoolUser[]","name":"","type":"tuple[]"},{"internalType":"uint256","name":"","type":"uint256"},{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"account","type":"address"}],"name":"getPoolsBySupplier","outputs":[{"internalType":"uint256[]","name":"","type":"uint256[]"},{"components":[{"internalType":"string","name":"name","type":"string"},{"internalType":"address","name":"creator","type":"address"},{"internalType":"address","name":"comptroller","type":"address"},{"internalType":"uint256","name":"blockPosted","type":"uint256"},{"internalType":"uint256","name":"timestampPosted","type":"uint256"}],"internalType":"struct FusePoolDirectory.FusePool[]","name":"","type":"tuple[]"}],"stateMutability":"view","type":"function","constant":true},{"inputs":[{"internalType":"address","name":"account","type":"address"}],"name":"getPoolsBySupplierWithData","outputs":[{"internalType":"uint256[]","name":"","type":"uint256[]"},{"components":[{"internalType":"string","name":"name","type":"string"},{"internalType":"address","name":"creator","type":"address"},{"internalType":"address","name":"comptroller","type":"address"},{"internalType":"uint256","name":"blockPosted","type":"uint256"},{"internalType":"uint256","name":"timestampPosted","type":"uint256"}],"internalType":"struct FusePoolDirectory.FusePool[]","name":"","type":"tuple[]"},{"components":[{"internalType":"uint256","name":"totalSupply","type":"uint256"},{"internalType":"uint256","name":"totalBorrow","type":"uint256"},{"internalType":"address[]","name":"underlyingTokens","type":"address[]"},{"internalType":"string[]","name":"underlyingSymbols","type":"string[]"},{"internalType":"bool","name":"whitelistedAdmin","type":"bool"}],"internalType":"struct FusePoolLens.FusePoolData[]","name":"","type":"tuple[]"},{"internalType":"bool[]","name":"","type":"bool[]"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"account","type":"address"}],"name":"getUserSummary","outputs":[{"internalType":"uint256","name":"","type":"uint256"},{"internalType":"uint256","name":"","type":"uint256"},{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"contract Comptroller","name":"comptroller","type":"address"},{"internalType":"address","name":"account","type":"address"}],"name":"getPoolUserSummary","outputs":[{"internalType":"uint256","name":"","type":"uint256"},{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"account","type":"address"}],"name":"getWhitelistedPoolsByAccount","outputs":[{"internalType":"uint256[]","name":"","type":"uint256[]"},{"components":[{"internalType":"string","name":"name","type":"string"},{"internalType":"address","name":"creator","type":"address"},{"internalType":"address","name":"comptroller","type":"address"},{"internalType":"uint256","name":"blockPosted","type":"uint256"},{"internalType":"uint256","name":"timestampPosted","type":"uint256"}],"internalType":"struct FusePoolDirectory.FusePool[]","name":"","type":"tuple[]"}],"stateMutability":"view","type":"function","constant":true},{"inputs":[{"internalType":"address","name":"account","type":"address"}],"name":"getWhitelistedPoolsByAccountWithData","outputs":[{"internalType":"uint256[]","name":"","type":"uint256[]"},{"components":[{"internalType":"string","name":"name","type":"string"},{"internalType":"address","name":"creator","type":"address"},{"internalType":"address","name":"comptroller","type":"address"},{"internalType":"uint256","name":"blockPosted","type":"uint256"},{"internalType":"uint256","name":"timestampPosted","type":"uint256"}],"internalType":"struct FusePoolDirectory.FusePool[]","name":"","type":"tuple[]"},{"components":[{"internalType":"uint256","name":"totalSupply","type":"uint256"},{"internalType":"uint256","name":"totalBorrow","type":"uint256"},{"internalType":"address[]","name":"underlyingTokens","type":"address[]"},{"internalType":"string[]","name":"underlyingSymbols","type":"string[]"},{"internalType":"bool","name":"whitelistedAdmin","type":"bool"}],"internalType":"struct FusePoolLens.FusePoolData[]","name":"","type":"tuple[]"},{"internalType":"bool[]","name":"","type":"bool[]"}],"stateMutability":"nonpayable","type":"function"}]"""
raw_secondary_abi = """[{"inputs":[],"name":"directory","outputs":[{"internalType":"contract FusePoolDirectory","name":"","type":"address"}],"stateMutability":"view","type":"function","constant":true},{"inputs":[{"internalType":"contract FusePoolDirectory","name":"_directory","type":"address"}],"name":"initialize","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"contract Comptroller","name":"comptroller","type":"address"}],"name":"getPoolOwnership","outputs":[{"internalType":"address","name":"","type":"address"},{"internalType":"bool","name":"","type":"bool"},{"internalType":"bool","name":"","type":"bool"},{"components":[{"internalType":"address","name":"cToken","type":"address"},{"internalType":"address","name":"admin","type":"address"},{"internalType":"bool","name":"adminHasRights","type":"bool"},{"internalType":"bool","name":"fuseAdminHasRights","type":"bool"}],"internalType":"struct FusePoolLensSecondary.CTokenOwnership[]","name":"","type":"tuple[]"}],"stateMutability":"view","type":"function","constant":true},{"inputs":[{"internalType":"address","name":"account","type":"address"},{"internalType":"contract CToken","name":"cTokenModify","type":"address"}],"name":"getMaxRedeem","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"account","type":"address"},{"internalType":"contract CToken","name":"cTokenModify","type":"address"}],"name":"getMaxBorrow","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"contract Comptroller","name":"comptroller","type":"address"}],"name":"getRewardSpeedsByPool","outputs":[{"internalType":"contract CToken[]","name":"","type":"address[]"},{"internalType":"contract RewardsDistributor[]","name":"","type":"address[]"},{"internalType":"address[]","name":"","type":"address[]"},{"internalType":"uint256[][]","name":"","type":"uint256[][]"},{"internalType":"uint256[][]","name":"","type":"uint256[][]"}],"stateMutability":"view","type":"function","constant":true},{"inputs":[{"internalType":"contract Comptroller[]","name":"comptrollers","type":"address[]"}],"name":"getRewardSpeedsByPools","outputs":[{"internalType":"contract CToken[][]","name":"","type":"address[][]"},{"internalType":"contract RewardsDistributor[][]","name":"","type":"address[][]"},{"internalType":"address[][]","name":"","type":"address[][]"},{"internalType":"uint256[][][]","name":"","type":"uint256[][][]"},{"internalType":"uint256[][][]","name":"","type":"uint256[][][]"}],"stateMutability":"view","type":"function","constant":true},{"inputs":[{"internalType":"address","name":"holder","type":"address"},{"internalType":"contract RewardsDistributor[]","name":"distributors","type":"address[]"}],"name":"getUnclaimedRewardsByDistributors","outputs":[{"internalType":"address[]","name":"","type":"address[]"},{"internalType":"uint256[]","name":"","type":"uint256[]"},{"internalType":"contract CToken[][]","name":"","type":"address[][]"},{"internalType":"uint256[2][][]","name":"","type":"uint256[2][][]"},{"internalType":"uint256[]","name":"","type":"uint256[]"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"supplier","type":"address"}],"name":"getRewardsDistributorsBySupplier","outputs":[{"internalType":"uint256[]","name":"","type":"uint256[]"},{"internalType":"contract Comptroller[]","name":"","type":"address[]"},{"internalType":"contract RewardsDistributor[][]","name":"","type":"address[][]"}],"stateMutability":"view","type":"function","constant":true}]"""


def get_primary_lens_abi() -> dict[str, spec.FunctionABI | spec.EventABI]:
    return {item['name']: item for item in json.loads(raw_primary_abi)}


def get_secondary_lens_abi() -> dict[str, spec.FunctionABI | spec.EventABI]:
    return {item['name']: item for item in json.loads(raw_secondary_abi)}


@functools.lru_cache()
def get_lens_abi() -> dict[str, spec.FunctionABI | spec.EventABI]:
    return dict(
        get_primary_lens_abi(),
        **get_secondary_lens_abi(),
    )


def get_function_abi(
    function_name: str, **abi_query: typing.Any
) -> spec.FunctionABI:
    return evm.get_function_abi(
        function_name=function_name,
        contract_abi=list(get_lens_abi().values()),
        **abi_query,
    )



from __future__ import annotations

import toolcli

from ctc.protocols import fourbyte_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_fourbyte_build_command,
        'help': 'build local copy of 4byte database\n\nscrapes from 4byte.directory',
        'args': [
            {
                'name': 'datatype',
                'choices': ['functions', 'events'],
                'nargs': '?',
                'help': '"functions" or "events", omit to build database of both',
            },
        ],
        'examples': {
            'functions': {
                'description': 'build function signature db',
                'long': True,
            },
            'events': {'description': 'build event signature db', 'long': True},
        },
    }


async def async_fourbyte_build_command(datatype: str | None) -> None:
    if datatype is None:
        datatypes = ['functions', 'events']
    else:
        datatypes = [datatype]

    print('Building local copy of 4byte database...')

    if 'functions' in datatypes:
        print('Building function signature database...')
        await fourbyte_utils.async_build_function_signatures_dataset()

    if 'events' in datatypes:
        print('Building event signature database...')
        await fourbyte_utils.async_build_event_signatures_dataset()

    print()
    print('local 4byte database up to date')



from __future__ import annotations

import typing

from ctc import spec
from . import format_utils

if typing.TYPE_CHECKING:
    from typing_extensions import Literal


@typing.overload
def keccak(
    data: spec.GenericBinaryData,
    output_format: Literal['integer'],
    *,
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> int:
    ...


@typing.overload
def keccak(
    data: spec.GenericBinaryData,
    output_format: Literal['binary'],
    *,
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> bytes:
    ...


@typing.overload
def keccak(
    data: spec.GenericBinaryData,
    output_format: Literal['prefix_hex', 'raw_hex', None] = 'prefix_hex',
    *,
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> str:
    ...


def keccak(
    data: spec.GenericBinaryData,
    output_format: typing.Optional[spec.BinaryFormat] = 'prefix_hex',
    *,
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> spec.GenericBinaryData:
    """return keccack-256 hash of hex or binary data"""

    # test that data is not a textual input
    if isinstance(data, str) and data != '0x':
        try:
            int(data, 16)
        except ValueError:
            raise Exception(
                'for text data, use keccak_text() instead of keccak()'
            )

    # determine library
    if library is None:
        try:
            import sha3  # type: ignore

            library = 'pysha3'

        except ImportError:
            library = 'pycryptodome'

    # convert data to binary
    data = format_utils.to_binary(data)

    if library == 'pysha3':
        import sha3

        binary = sha3.keccak_256(data).digest()
    elif library == 'pycryptodome':
        from Crypto.Hash import keccak as f_keccak

        binary = f_keccak.new(digest_bits=256, data=data).digest()
    else:
        raise Exception(
            'must choose valid library, either \'pysha3\' or \'pycryptodome\''
        )

    return format_utils.binary_convert(binary, output_format)


@typing.overload
def keccak_text(
    text: typing.Union[str, bytes],
    output_format: Literal['integer'],
    *,
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> int:
    ...


@typing.overload
def keccak_text(
    text: typing.Union[str, bytes],
    output_format: Literal['binary'],
    *,
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> bytes:
    ...


@typing.overload
def keccak_text(
    text: typing.Union[str, bytes],
    output_format: Literal['prefix_hex', 'raw_hex'] = 'prefix_hex',
    *,
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> str:
    ...


def keccak_text(
    text: typing.Union[str, bytes],
    output_format: spec.BinaryFormat = 'prefix_hex',
    *,
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> spec.GenericBinaryData:
    """return keccack-256 hash of text"""

    if isinstance(text, str):
        text = text.encode()

    return keccak(text, output_format=output_format, library=library)



from __future__ import annotations

import asyncio
import typing

from ctc import evm
from ctc import spec

if typing.TYPE_CHECKING:
    from typing_extensions import Literal

    import tooltime


class DEX:
    """Standardized interface for DEXes

    - All methods should be classmethods or staticmethods
    - The class should never be instantiated
    - Users do not need to use DEX classes directly
        - Functions in dex_utils will call the appropriate DEX methods

    Subclasses should implement the following methods:
    - async_get_new_pools()
    - _async_get_pool_assets_from_node()
    - _async_get_pool_raw_trades()

    Subclasses should specify the following properties:
    - _pool_factories
    """

    _pool_factories: typing.Mapping[
        int, typing.Sequence[spec.Address]
    ] | None = None

    def __init__(self) -> None:
        raise Exception(
            'this class should never be initialized'
            '\n\nuse functions in dex_utils instead'
        )

    #
    # # methods needing subclass implementation
    #

    @classmethod
    async def async_get_new_pools(
        cls,
        *,
        factory: spec.Address,
        start_block: spec.BlockNumberReference | None = None,
        end_block: spec.BlockNumberReference | None = None,
        start_time: tooltime.Timestamp | None = None,
        end_time: tooltime.Timestamp | None = None,
        context: spec.Context = None,
    ) -> typing.Sequence[spec.DexPool]:
        raise NotImplementedError(cls.__name__ + '.async_get_new_pools')

    @classmethod
    async def _async_get_pool_assets_from_node(
        cls,
        pool: spec.Address,
        *,
        block: spec.BlockNumberReference | None = None,
        context: spec.Context = None,
    ) -> typing.Sequence[spec.Address]:
        raise NotImplementedError(cls.__name__ + '.async_get_pool_assets')

    @classmethod
    async def _async_get_pool_raw_trades(
        cls,
        pool: spec.Address,
        *,
        start_block: spec.BlockNumberReference | None = None,
        end_block: spec.BlockNumberReference | None = None,
        start_time: tooltime.Timestamp | None = None,
        end_time: tooltime.Timestamp | None = None,
        include_timestamps: bool = False,
        verbose: bool = False,
        context: spec.Context = None,
    ) -> spec.RawDexTrades:
        raise NotImplementedError(cls.__name__ + '.async_get_pool_trades')

    #
    # # dex metadata
    #

    @classmethod
    def get_dex_name(cls) -> str:
        return cls.__name__.rstrip('DEX')

    @classmethod
    def get_pool_factories(
        cls,
        context: spec.Context = None,
    ) -> typing.Sequence[spec.Address]:
        from ctc import config

        chain_id = config.get_context_chain_id(context)
        if cls._pool_factories is None:
            raise NotImplementedError(cls.__name__ + '._pool_factories')
        if chain_id not in cls._pool_factories:
            raise Exception('unknown network: ' + str(chain_id))
        return cls._pool_factories[chain_id]

    #
    # # multiple pool functions
    #

    @classmethod
    async def async_get_pools(
        cls,
        *,
        factory: spec.Address | None = None,
        all_dex_factories: bool = False,
        assets: typing.Sequence[spec.Address] | None = None,
        start_block: spec.BlockNumberReference | None = None,
        end_block: spec.BlockNumberReference | None = None,
        start_time: tooltime.Timestamp | None = None,
        end_time: tooltime.Timestamp | None = None,
        update: bool = False,
        context: spec.Context = None,
    ) -> typing.Sequence[spec.DexPool]:
        """return pools"""

        factories: typing.Sequence[spec.Address] | None = None
        if all_dex_factories:
            pool_factories = cls.get_pool_factories(context=context)
            if len(pool_factories) == 1:
                factory = pool_factories[0]
                factories = None
            elif len(pool_factories) > 1:
                factory = None
                factories = pool_factories
            else:
                factory = None
                factories = None

        if assets is not None and len(assets) == 0:
            assets = None

        start_block, end_block = await evm.async_resolve_block_range(
            start_block=start_block,
            end_block=end_block,
            start_time=start_time,
            end_time=end_time,
            allow_none=True,
            context=context,
        )

        if start_block is not None:
            start_block = await evm.async_block_number_to_int(
                start_block,
                context=context,
            )
        if end_block is not None:
            end_block = await evm.async_block_number_to_int(
                end_block,
                context=context,
            )

        pools = await cls.async_get_stored_dex_pools(
            factory=factory,
            factories=factories,
            assets=assets,
            start_block=start_block,
            end_block=end_block,
            context=context,
        )

        if update:
            new_pools = await cls.async_update_pools(
                factory=factory,
                factories=factories,
                context=context,
            )

            # filter pools according to function inputs
            new_pools = _filter_pools(
                pools=new_pools,
                assets=assets,
                start_block=start_block,
                end_block=end_block,
            )

            pools = list(pools) + list(new_pools)

        return pools

    @classmethod
    async def async_get_stored_dex_pools(
        cls,
        *,
        factory: spec.Address | None = None,
        factories: typing.Sequence[spec.Address] | None = None,
        assets: typing.Sequence[spec.Address] | None = None,
        start_block: spec.BlockNumberReference | None = None,
        end_block: spec.BlockNumberReference | None = None,
        context: spec.Context = None,
    ) -> typing.Sequence[spec.DexPool]:
        """return pools"""

        from ctc import db

        pools = await db.async_query_dex_pools(
            factory=factory,
            factories=factories,
            assets=assets,
            start_block=start_block,
            end_block=end_block,
            context=context,
        )
        if pools is None:
            pools = []

        return pools

    #
    # # multiple pool updates
    #

    @classmethod
    async def async_update_pools(
        cls,
        *,
        factory: spec.Address | None = None,
        factories: typing.Sequence[spec.Address] | None = None,
        context: spec.Context = None,
    ) -> typing.Sequence[spec.DexPool]:
        """update latest pools of factory"""

        from ctc import db

        # if not factory or factories specified, use all
        if factories is None and factory is None:
            factories = cls.get_pool_factories(context=context)

        # if using multiple factories, call function separately for each
        if factories is not None:
            coroutines = []
            for factory in factories:
                coroutine = cls.async_update_pools(
                    factory=factory,
                    context=context,
                )
                coroutines.append(coroutine)
            results = await asyncio.gather(*coroutines)
            return [dex_pool for result in results for dex_pool in result]

        if factory is None:
            raise Exception('factory should be specified')

        # get new pools
        last_scanned_block = (
            await db.async_query_dex_pool_factory_last_scanned_block(
                factory=factory,
                context=context,
            )
        )
        if last_scanned_block is None:
            from ctc.toolbox import search_utils

            try:
                creation_block = await evm.async_get_contract_creation_block(
                    factory,
                    context=context,
                )
                if creation_block is None:
                    raise Exception(
                        'could not determine factory creation block'
                    )
                last_scanned_block = creation_block - 1
            except search_utils.NoMatchFound:
                last_scanned_block = -1

        latest_block = await evm.async_get_latest_block_number(context=context)

        if last_scanned_block + 1 <= latest_block:
            new_pools = await cls.async_get_new_pools(
                factory=factory,
                start_block=last_scanned_block + 1,
                end_block=latest_block,
                context=context,
            )
            await db.async_intake_dex_pools(
                factory=factory,
                dex_pools=new_pools,
                last_scanned_block=latest_block,
                context=context,
            )

        else:
            new_pools = []

        return new_pools

    #
    # # single pool metadata
    #

    @classmethod
    async def async_get_pool_assets(
        cls,
        pool: spec.Address,
        *,
        block: spec.BlockNumberReference | None = None,
        context: spec.Context = None,
    ) -> typing.Sequence[spec.Address]:
        return await cls._async_get_pool_assets_from_node(
            pool=pool,
            block=block,
            context=context,
        )

    @classmethod
    async def async_get_pool_asset_symbols(
        cls,
        pool: spec.Address,
        *,
        context: spec.Context = None,
    ) -> typing.Sequence[str]:
        assets = await cls.async_get_pool_assets(pool=pool, context=context)
        return await evm.async_get_erc20s_symbols(assets, context=context)

    #
    # # single pool balances
    #

    @classmethod
    async def async_get_pool_balance(
        cls,
        pool: spec.Address,
        asset: spec.Address,
        *,
        factory: spec.Address | None = None,
        normalize: bool = True,
        block: spec.BlockNumberReference | None = None,
        context: spec.Context = None,
    ) -> int | float:
        result = await evm.async_get_erc20_balance(
            wallet=pool,
            token=asset,
            normalize=normalize,
            block=block,
            context=context,
        )
        if result is None:
            raise Exception('invalid result for pool balance')
        return result

    @classmethod
    async def async_get_pool_balances(
        cls,
        pool: spec.Address,
        *,
        factory: spec.Address | None = None,
        normalize: bool = True,
        block: spec.BlockNumberReference | None = None,
        context: spec.Context = None,
    ) -> typing.Mapping[spec.Address, int | float]:
        pool_tokens = await cls.async_get_pool_assets(
            pool=pool, context=context
        )

        if cls.async_get_pool_balance == DEX.async_get_pool_balance:
            balances = await evm.async_get_erc20s_balances(
                wallet=pool,
                tokens=pool_tokens,
                normalize=normalize,
                block=block,
                context=context,
            )
            return dict(zip(pool_tokens, balances))
        else:
            coroutines = []
            for token in pool_tokens:
                coroutine = cls.async_get_pool_balance(
                    pool=pool,
                    asset=token,
                    normalize=normalize,
                    block=block,
                    context=context,
                )
                coroutines.append(coroutine)
            results = await asyncio.gather(*coroutines)
            return dict(zip(pool_tokens, results))

    @classmethod
    async def async_get_pool_balance_by_block(
        cls,
        pool: spec.Address,
        asset: spec.Address,
        *,
        blocks: typing.Sequence[spec.BlockNumberReference],
        factory: spec.Address | None = None,
        normalize: bool = True,
        context: spec.Context = None,
    ) -> typing.Sequence[int | float]:
        return await evm.async_get_erc20_balance_by_block(
            wallet=pool,
            token=asset,
            normalize=normalize,
            blocks=blocks,
            context=context,
        )

    @classmethod
    async def async_get_pool_balances_by_block(
        cls,
        pool: spec.Address,
        *,
        blocks: typing.Sequence[spec.BlockNumberReference],
        factory: spec.Address | None = None,
        normalize: bool = True,
        context: spec.Context = None,
    ) -> typing.Mapping[str, typing.Sequence[int | float]]:
        if len(blocks) == 0:
            raise NotImplementedError('must specify blocks')

        coroutines = [
            cls.async_get_pool_balances(
                pool=pool,
                factory=factory,
                normalize=normalize,
                block=block,
                context=context,
            )
            for block in blocks
        ]
        results = await asyncio.gather(*coroutines)
        balances_by_block: typing.MutableMapping[
            str, typing.MutableSequence[int | float]
        ] = {}
        for result in results:
            for asset, balance in result.items():
                balances_by_block.setdefault(asset, [])
                balances_by_block[asset].append(balance)

        return balances_by_block

    #
    # # single pool swaps functions
    #

    @classmethod
    async def async_get_pool_trades(
        cls,
        pool: spec.Address,
        *,
        start_block: spec.BlockNumberReference | None = None,
        end_block: spec.BlockNumberReference | None = None,
        start_time: tooltime.Timestamp | None = None,
        end_time: tooltime.Timestamp | None = None,
        label: Literal['index', 'symbol', 'address'] = 'index',
        include_timestamps: bool = False,
        normalize: bool = True,
        verbose: bool = False,
        remove_missing_fields: bool = True,
        include_prices: bool = False,
        include_volumes: bool = False,
        context: spec.Context = None,
    ) -> spec.DataFrame:
        import polars as pl

        # queue relevant label data
        if label == 'symbol':
            symbols_coroutine = cls.async_get_pool_asset_symbols(
                pool=pool,
                context=context,
            )
            symbols_task = asyncio.create_task(symbols_coroutine)
        if normalize or label == 'address':
            assets_coroutine = cls.async_get_pool_assets(
                pool=pool,
                context=context,
            )
            assets_task = asyncio.create_task(assets_coroutine)

        output = await cls._async_get_pool_raw_trades(
            pool=pool,
            start_block=start_block,
            end_block=end_block,
            start_time=start_time,
            end_time=end_time,
            include_timestamps=include_timestamps,
            verbose=verbose,
            context=context,
        )

        if remove_missing_fields:
            if 'timestamp' in output and output['timestamp'] is None:
                del output['timestamp']
            if 'recipient' in output and output['recipient'] is None:
                del output['recipient']

        if normalize or label == 'address':
            assets = await assets_task

        # normalize
        if normalize:
            # test for metapools
            if max(output['sold_id']) >= len(assets) or max(
                output['bought_id']
            ) >= len(assets):
                raise NotImplementedError(
                    'normalize not implemented for metapools'
                )

            decimals = await evm.async_get_erc20s_decimals(
                assets,
                context=context,
            )

            sold_decimals = output['sold_id'].apply(lambda i: decimals[i])
            bought_decimals = output['bought_id'].apply(lambda i: decimals[i])
            output['sold_amount'] /= 10**sold_decimals
            output['bought_amount'] /= 10**bought_decimals

        # replace labels
        if label in ['symbol', 'address']:
            if label == 'symbol':
                new_ids = await symbols_task
            elif label == 'address':
                new_ids = assets
            else:
                raise Exception('unknown label format: ' + str(label))

            output['bought_id'] = output['bought_id'].apply(
                lambda i: new_ids[i]
            )
            output['sold_id'] = output['sold_id'].apply(lambda i: new_ids[i])

        df = pl.DataFrame(output)

        if include_prices:
            prices = cls.compute_trade_prices(df, normalized=normalize)
            for key, value in prices.items():
                df = df.with_columns(pl.Series(key, value))
        if include_volumes:
            volumes = cls.compute_trade_volumes(df)
            for key, value in volumes.items():
                df = df.with_columns(pl.Series(key, value))

        return df

    @classmethod
    def compute_trade_volumes(
        cls, df: spec.DataFrame
    ) -> typing.Mapping[str, spec.Series]:
        df['sold_amount']

        all_ids = sorted(
            set(df['sold_id'].unique()) | set(df['bought_id'].unique())
        )

        volumes = {}
        for asset_id in all_ids:
            volumes['volume__' + str(asset_id)] = (
                df['sold_id'] == asset_id
            ) * df['sold_amount'] + (df['bought_id'] == asset_id) * df[
                'bought_amount'
            ]

        return volumes

    @classmethod
    def compute_trade_prices(
        cls,
        df: spec.DataFrame,
        normalized: bool,
    ) -> typing.Mapping[str, spec.Series]:
        import polars as pl

        if not normalized:
            raise Exception('including prices requires normalize=True')

        all_ids = sorted(
            set(df['sold_id'].unique()) | set(df['bought_id'].unique())
        )
        prices: typing.MutableMapping[str, spec.Series] = {}

        # pre-compute quantities
        sold_amount = df['sold_amount'].apply(float)
        bought_amount = df['bought_amount'].apply(float)
        sold_per_bought = sold_amount / bought_amount
        bought_per_sold = bought_amount / sold_amount
        sold_masks = {}
        bought_masks = {}
        for asset_id in all_ids:
            sold_masks[asset_id] = df['sold_id'] == asset_id
            bought_masks[asset_id] = df['bought_id'] == asset_id

        # compute combinations
        for lhs_id in all_ids:
            for rhs_id in all_ids:
                if lhs_id == rhs_id:
                    continue

                key = 'price__' + str(lhs_id) + '__per__' + str(rhs_id)

                sold_lhs_mask = sold_masks[lhs_id]
                sold_rhs_mask = sold_masks[rhs_id]
                bought_lhs_mask = bought_masks[lhs_id]
                bought_rhs_mask = bought_masks[rhs_id]

                values = (
                    sold_lhs_mask.cast(int)
                    * bought_rhs_mask.cast(int)
                    * sold_per_bought
                    + bought_lhs_mask.cast(int)
                    * sold_rhs_mask.cast(int)
                    * bought_per_sold
                )
                df = df.with_columns(
                    pl.Series(key, values)
                )
                if len(all_ids) > 2:
                    combined_mask = (
                        sold_lhs_mask * bought_rhs_mask
                        + bought_lhs_mask * sold_rhs_mask
                    )
                    df[key][~combined_mask] = float('nan')

                prices[key] = values

        return prices

    # should combine adds and removes into single function?
    # @classmethod
    # async def async_get_pool_liquidity_adds(
    #     cls,
    #     pool: spec.Address,
    #     *,
    #     start_block: spec.BlockNumberReference | None = None,
    #     end_block: spec.BlockNumberReference | None = None,
    #     start_time: tooltime.Timestamp | None = None,
    #     end_time: tooltime.Timestamp | None = None,
    #     labels: Literal['index', 'symbol', 'address'] = 'index',
    #     include_timestamps: bool = False,
    #     normalize: bool = True,
    #     network: spec.NetworkReference | None = None,
    #     provider: spec.ProviderReference | None = None,
    #     verbose: bool = False,
    # ) -> spec.DataFrame:
    #     raise NotImplementedError(cls.__name__ + '.async_get_pool_trades')

    # @classmethod
    # async def async_get_pool_liquidity_removes(
    #     cls,
    #     pool: spec.Address,
    #     *,
    #     start_block: spec.BlockNumberReference | None = None,
    #     end_block: spec.BlockNumberReference | None = None,
    #     start_time: tooltime.Timestamp | None = None,
    #     end_time: tooltime.Timestamp | None = None,
    #     labels: Literal['index', 'symbol', 'address'] = 'index',
    #     include_timestamps: bool = False,
    #     normalize: bool = True,
    #     network: spec.NetworkReference | None = None,
    #     provider: spec.ProviderReference | None = None,
    #     verbose: bool = False,
    # ) -> spec.DataFrame:
    #     raise NotImplementedError(cls.__name__ + '.async_get_pool_trades')


def _filter_pools(
    *,
    pools: typing.Sequence[spec.DexPool],
    assets: typing.Sequence[str] | None = None,
    start_block: int | None = None,
    end_block: int | None = None,
) -> typing.Sequence[spec.DexPool]:
    filtered = []

    # filter the new pools according to input arguments
    for pool in pools:
        # check asset filter
        include = True
        if assets is not None:
            keys = ['asset0', 'asset1', 'asset2', 'asset3']
            for asset in assets:
                asset = asset.lower()
                include = any(pool[key] == asset for key in keys)  # type: ignore
                if not include:
                    break
        if not include:
            continue

        # check block range
        if start_block is not None and (
            pool['creation_block'] is None
            or pool['creation_block'] < start_block
        ):
            continue
        if end_block is not None and (
            pool['creation_block'] is None or pool['creation_block'] > end_block
        ):
            continue

        filtered.append(pool)

    return filtered




from __future__ import annotations

import typing

from ctc import spec
from .. import abi_utils
from .. import block_utils
from .. import contract_utils
from . import event_metadata
from . import event_query_utils

if typing.TYPE_CHECKING:
    import tooltime

    from typing_extensions import Literal


async def async_get_events(
    contract_address: spec.Address | None = None,
    *,
    event_name: str | None = None,
    event_abi: spec.EventABI | None = None,
    event_hash: str | None = None,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    context: spec.Context = None,
    named_topics: typing.Mapping[str, typing.Any] | None = None,
    topic1: typing.Any | None = None,
    topic2: typing.Any | None = None,
    topic3: typing.Any | None = None,
    topic1_is_binary: bool | None = None,
    topic2_is_binary: bool | None = None,
    topic3_is_binary: bool | None = None,
    verbose: int | bool = 1,
    decode: bool = True,
    share_abis_across_contracts: bool = True,
    include_timestamps: bool = False,
    include_event_names: bool = False,
    binary_output_format: Literal['binary', 'prefix_hex'] = 'prefix_hex',
    integer_output_format: spec.IntegerOutputFormat | None = None,
    only_columns: typing.Sequence[str] | None = None,
    exclude_columns: typing.Sequence[str] | None = None,
    convert_ints: type | bool = False,
    convert_invalid_str_to_none: bool = False,
    convert_invalid_str_to: None | str = None,
    max_blocks_per_request: int = 2000,
) -> spec.DataFrame:
    """get events"""

    from . import event_hybrid_queries

    # get query inputs
    (
        start_block,
        end_block,
        encoded_topics,
        event_abi,
        columns_to_load,
    ) = await _async_get_query_inputs(
        contract_address=contract_address,
        event_name=event_name,
        event_abi=event_abi,
        event_hash=event_hash,
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        context=context,
        named_topics=named_topics,
        topic1=topic1,
        topic2=topic2,
        topic3=topic3,
        topic1_is_binary=topic1_is_binary,
        topic2_is_binary=topic2_is_binary,
        topic3_is_binary=topic3_is_binary,
        decode=decode,
        only_columns=only_columns,
        exclude_columns=exclude_columns,
        include_event_names=include_event_names,
    )

    # query data from db and/or node
    df = await event_hybrid_queries._async_query_events_from_node_and_db(
        contract_address=contract_address,
        event_hash=encoded_topics[0],
        topic1=encoded_topics[1],
        topic2=encoded_topics[2],
        topic3=encoded_topics[3],
        start_block=start_block,
        end_block=end_block,
        verbose=verbose,
        binary_output_format=binary_output_format,
        columns_to_load=columns_to_load,
        context=context,
        max_blocks_per_request=max_blocks_per_request,
    )

    # post-process result
    return await _async_postprocess_query_result(
        df=df,
        event_abi=event_abi,
        verbose=bool(verbose),
        columns_to_load=columns_to_load,
        decode=decode,
        context=context,
        share_abis_across_contracts=share_abis_across_contracts,
        include_timestamps=include_timestamps,
        include_event_names=include_event_names,
        binary_output_format=binary_output_format,
        integer_output_format=integer_output_format,
        convert_invalid_str_to_none=convert_invalid_str_to_none,
        convert_invalid_str_to=convert_invalid_str_to,
        convert_ints=convert_ints,
    )


async def _async_postprocess_query_result(
    *,
    df: spec.DataFrame,
    event_abi: spec.EventABI | None,
    verbose: bool,
    columns_to_load: typing.Sequence[str],
    decode: bool = True,
    context: spec.Context,
    share_abis_across_contracts: bool = True,
    include_timestamps: bool = False,
    include_event_names: bool = False,
    binary_output_format: Literal['binary', 'prefix_hex'] = 'prefix_hex',
    integer_output_format: spec.IntegerOutputFormat | None = None,
    convert_invalid_str_to_none: bool = False,
    convert_invalid_str_to: None | str = None,
    convert_ints: type | bool = False,
) -> spec.DataFrame:

    import polars as pl

    # summarize output
    if verbose >= 2:
        from ctc import cli

        print()
        print('events gathered')
        cli.print_bullet(key='n_events', value=len(df))
        if 'contract_address' in columns_to_load:
            n_contracts = df['contract_address'].n_unique()
            cli.print_bullet(key='n_contracts', value=n_contracts)
        if 'event_hash' in columns_to_load:
            n_event_types = df['event_hash'].n_unique()
            cli.print_bullet(key='n_event_types', value=n_event_types)

    # insert metadata columns
    if include_timestamps:

        timestamps = await event_metadata.async_get_event_timestamps(
            df,
            context=context,
        )
        df = df.with_columns(pl.Series(name='timestamp', values=timestamps))
    if include_event_names:
        event_names = await event_metadata._async_get_event_names_column(
            events=df,
            share_abis_across_contracts=share_abis_across_contracts,
            context=context,
        )
        df = df.with_columns(pl.Series(name='event_name', values=event_names))

    # format data
    if decode:
        decoded = await abi_utils.async_decode_events_dataframe(
            df,
            event_abis=([event_abi] if event_abi is not None else None),
            context=context,
            binary_output_format=binary_output_format,
            integer_output_format=integer_output_format,
            convert_invalid_str_to_none=convert_invalid_str_to_none,
            convert_invalid_str_to=convert_invalid_str_to,
        )
        df = df.with_columns(decoded)
        encoded_columns = ['topic1', 'topic2', 'topic3', 'unindexed']
        to_drop = [column for column in encoded_columns if column in df]
        df = df.drop(to_drop)

    if convert_ints:
        int_columns = df.select(pl.col(pl.Object)).columns
        if isinstance(convert_ints, bool):
            import numpy as np

            converted = []
            for column in int_columns:
                as_array = np.array(df[column].to_list())
                max_value = as_array.max()
                min_value = as_array.min()
                if max_value > 2 ** 32 - 1 and min_value < 2 ** 32 + 1:
                    as_array = as_array.astype(float)
                    dtype: pl.datatypes.DataTypeClass = pl.Float64
                else:
                    dtype = pl.Int64
                converted.append(pl.Series(column, as_array, dtype=dtype))
            df = df.with_columns(converted)
        elif isinstance(convert_ints, pl.datatypes.DataTypeClass):
            converted = []
            for column in int_columns:
                new_column = pl.Series(column, df[column].to_list(), dtype=convert_ints)
                converted.append(new_column)
            df = df.with_columns(converted)
        elif isinstance(convert_ints, type):
            return_dtypes = {
                int: pl.datatypes.Int64,
                float: pl.datatypes.Float64,
            }
            return_dtype = return_dtypes.get(convert_ints)
            converted = [
                df[column].apply(convert_ints, return_dtype=return_dtype)
                for column in int_columns
            ]
            df = df.with_columns(converted)
        else:
            raise Exception('unknown format for convert ints')

    return df


async def _async_get_query_inputs(
    contract_address: spec.Address | None = None,
    *,
    event_name: str | None = None,
    event_abi: spec.EventABI | None = None,
    event_hash: str | None = None,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    context: spec.Context = None,
    named_topics: typing.Mapping[str, typing.Any] | None = None,
    topic1: typing.Any | None = None,
    topic2: typing.Any | None = None,
    topic3: typing.Any | None = None,
    topic1_is_binary: bool | None = None,
    topic2_is_binary: bool | None = None,
    topic3_is_binary: bool | None = None,
    decode: bool = True,
    only_columns: typing.Sequence[str] | None = None,
    exclude_columns: typing.Sequence[str] | None = None,
    include_event_names: bool = False,
) -> tuple[
    int, int, typing.Sequence[bytes | None], spec.EventABI | None, typing.Sequence[str]
]:

    # determine start and end block
    start_block, end_block = await block_utils.async_resolve_block_range(
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        allow_none=True,
        to_int=True,
        end_none_means='latest',
        context=context,
    )
    if start_block is None:
        if contract_address is None:
            raise Exception('must specify contract_address or start_block')
        start_block = await contract_utils.async_get_contract_creation_block(
            contract_address=contract_address,
            context=context,
        )
    if start_block is None:
        raise Exception('could not determine start_block')
    if end_block is None:
        raise Exception('could not determine end_block')

    # get event query metadata
    (
        encoded_topics,
        event_abi,
    ) = await event_query_utils._async_parse_event_query_args(
        contract_address=contract_address,
        event_name=event_name,
        event_hash=event_hash,
        event_abi=event_abi,
        named_topics=named_topics,
        topic1=topic1,
        topic2=topic2,
        topic3=topic3,
        topic1_is_binary=topic1_is_binary,
        topic2_is_binary=topic2_is_binary,
        topic3_is_binary=topic3_is_binary,
        context=context,
    )

    # compute which columns to return
    columns_to_load = _get_columns_to_load(
        decode=decode,
        event_abi=event_abi,
        only_columns=only_columns,
        exclude_columns=exclude_columns,
        include_event_names=include_event_names,
    )

    return (
        start_block,
        end_block,
        encoded_topics,
        event_abi,
        columns_to_load,
    )


def _get_columns_to_load(
    *,
    only_columns: typing.Sequence[str] | None,
    exclude_columns: typing.Sequence[str] | None,
    decode: bool,
    event_abi: spec.EventABI | None,
    include_event_names: bool,
) -> typing.Sequence[str]:
    """specify which columns to use for partial loading

    need to account for dependencies
    - .e.g. if want to decode an indexed topic, need to load raw binary column
    """

    if only_columns is None and exclude_columns is None:
        exclude_columns = []
    if only_columns is not None and exclude_columns is not None:
        raise Exception('can only specify only_columns or exclude_columns')

    def adding(column: str) -> bool:
        return (only_columns is not None and column in only_columns) or (
            exclude_columns is not None and column not in exclude_columns
        )

    # index columns
    columns = []
    for column in [
        'block_number',
        'transaction_index',
        'log_index',
        'transaction_hash',
        'contract_address',
        'event_hash',
    ]:
        if adding(column):
            columns.append(column)

    if decode:
        # fetch topic and data columns if need to decode topics

        if event_abi is None:
            for column in ['topic1', 'topic2', 'topic3', 'unindexed']:
                if adding(column):
                    columns.append(column)

        else:

            # indexed names
            indexed = abi_utils.get_event_indexed_names(event_abi)
            n_indexed = len(indexed)
            if n_indexed >= 1:
                name = 'arg__' + indexed[0]
                if adding(name):
                    columns.append('topic1')
            if n_indexed >= 2:
                name = 'arg__' + indexed[1]
                if adding(name):
                    columns.append('topic2')
            if n_indexed >= 3:
                name = 'arg__' + indexed[2]
                if adding(name):
                    columns.append('topic3')

            # unindexed names
            unindexed = abi_utils.get_event_unindexed_names(event_abi)
            if any(adding('arg__' + name) for name in unindexed):
                columns.append('unindexed')

    else:
        # fetch topic and data columns if emitted by abi

        if event_abi is not None:
            n_indexed = len(abi_utils.get_event_indexed_names(event_abi))
            n_unindexed = len(abi_utils.get_event_unindexed_names(event_abi))
        else:
            n_indexed = 3
            n_unindexed = 1

        if n_indexed >= 1 and adding('topic1'):
            columns.append('topic1')
        if n_indexed >= 2 and adding('topic2'):
            columns.append('topic2')
        if n_indexed >= 3 and adding('topic3'):
            columns.append('topic3')
        if n_unindexed >= 1 and adding('unindexed'):
            columns.append('unindexed')

    # validate
    if only_columns is not None:

        # control specialty columns elsewhere
        if 'timestamp' in only_columns:
            raise Exception(
                'use include_timestamps to include/exclude timestamp'
            )
        if 'event_name' in only_columns:
            raise Exception(
                'use include_event_names to include/exclude timestamp'
            )
        for column in columns:
            if column in {
                'block_number',
                'transaction_index',
                'log_index',
            }:
                # ok, this is expected
                pass
            elif column not in only_columns:
                if column.startswith('arg__') and not decode:
                    message = 'cannot include arg__ columns when decode=False'
                elif (
                    column in {'topic1', 'topic2', 'topic3', 'unindexed'}
                    and decode
                ):
                    continue
                else:
                    message = 'cannot include column: ' + str(column)
                raise Exception(message)
    if exclude_columns is not None:
        if 'timestamp' in exclude_columns:
            raise Exception(
                'use include_timestamps to include/exclude timestamp'
            )
        if 'event_name' in exclude_columns:
            raise Exception(
                'use include_event_names to include/exclude timestamp'
            )
        for column in exclude_columns:
            if column in columns:
                if column == 'block_number':
                    message = 'cannot exclude block_number'
                else:
                    message = 'cannot exclude column: ' + str(column)
                raise Exception(message)
    if include_event_names and (
        'event_hash' not in columns or 'contract_address' not in columns
    ):
        raise Exception(
            'must load event_hash and contract_address to load event names'
        )

    return columns

